import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.regression_mae import RegressionMAE


class RegressionMAECPU(RegressionMAE[np.ndarray], MetricCPU):

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        # return np.sum(np.absolute(y_targ - y_pred))
        diff = np.subtract(y_targ, y_pred, dtype=self.model.dtype)
        np.absolute(diff, out=diff, dtype=self.model.dtype, casting="unsafe")
        return np.mean(diff, dtype=self.model.dtype)



import inspect
from unittest import SkipTest

import numpy as np

from pydtnn.libs.libconvdirect import ConvDirect
from pydtnn.backends.cpu.utils.im2row_nhwc_cython import im2row_nhwc_cython
from pydtnn.tests.abstract.common import verbose_test
from pydtnn.tests.abstract.conv_common import ConvCommonTestCase
from pydtnn.utils import print_with_header


class ConvDirectTestCase(ConvCommonTestCase):
    """
    Tests that conv_direct leads to the same results as i2c and mm.
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global ConvCommonTestCase
    del ConvCommonTestCase

    @classmethod
    def _compute_both(cls, weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None,
                      vpadding=0, hpadding=0, vstride=1, hstride=1,
                      vdilation=1, hdilation=1) -> tuple[np.ndarray, np.ndarray]:
        c, kh, kw, kn = weights.shape
        # b, c, h, w = x.shape
        cg_biases = biases.copy() if biases is not None else None
        conv_direct_result: np.ndarray = cls._compute(weights, x, biases=cg_biases,
                                                      kw=kw, kh=kh,
                                                      vpadding=vpadding, hpadding=hpadding,
                                                      vstride=vstride, hstride=hstride,
                                                      vdilation=vdilation, hdilation=hdilation)
        conv_direct_result: np.ndarray = conv_direct_result.reshape((-1, kn), copy=False)

        n, h, w, c = x.shape

        ho = (h + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
        wo = (w + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1

        dim_n = n * ho * wo
        dim_c = c * kh * kw

        x_c: np.ndarray = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

        im2row_nhwc_cython(x, x_c,
                           kh, kw, ho, wo,
                           vpadding, hpadding,
                           vstride, hstride,
                           vdilation, hdilation)
        w_c = weights.reshape((-1, kn), copy=False)
        im2row_mm_result: np.ndarray = np.matmul(x_c, w_c)
        if biases is not None:
            np.add(im2row_mm_result, biases.reshape((-1, kn), copy=False), out=im2row_mm_result, dtype=im2row_mm_result.dtype)
        if verbose_test():
            print_with_header("{} conv_direct_result".format(inspect.stack()[1][3]), conv_direct_result)
            print("Shape: ", conv_direct_result.shape,
                  " Sum: ", conv_direct_result.sum(),
                  " Min: ", conv_direct_result.min(),
                  " Max: ", conv_direct_result.max())
            print_with_header("{} im2row_mm_result".format(inspect.stack()[1][3]), im2row_mm_result)
            print("Shape: ", im2row_mm_result.shape,
                  " Sum: ", im2row_mm_result.sum(),
                  " Min: ", im2row_mm_result.min(),
                  " Max: ", im2row_mm_result.max())
            print("---")
            print("Maximum difference: ",
                  max([abs(x - y) for x, y in zip(conv_direct_result.flatten(), im2row_mm_result.flatten())]))
            print("---")
        return conv_direct_result, im2row_mm_result

    @staticmethod
    def _compute(weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None, kh=1, kw=1, vpadding=0, hpadding=0, vstride=1, hstride=1, vdilation=1, hdilation=1):
        if biases is not None:
            raise SkipTest("Direct does not support biases")
        return ConvDirect(method_name="convdirect_original_nhwc_default", debug=False).conv_direct(weights, x, None, vpadding, hpadding, vstride, hstride, vdilation, hdilation)



import numpy as np
from pydtnn.backends.cpu.utils.log_activation_cython import log_bwd_cython, log_fwd_cython

from pydtnn.activations.log import Log
from pydtnn.backends.cpu.activations.activation import ActivationCPU


class LogCPU(Log[np.ndarray], ActivationCPU):

    def initialize(self, prev_shape, x=None):
        super().initialize(prev_shape, x)
        # NOTE: These attributes only store data, their value before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self.y = np.zeros(shape=(self.model.batch_size, *self.shape), dtype=self.model.dtype, order="C")
        self.dx = np.zeros(shape=(self.model.batch_size, *self.shape), dtype=self.model.dtype, order="C")

    def _forward_numpy(self, x: np.ndarray) -> np.ndarray:
        # def forward(self, x: np.ndarray) -> np.ndarray:
        y = self.y[:x.shape[0], :]
        # y = np.log(1 / (1 + np.exp(-x)))
        np.multiply(x, -1, out=x,
                    dtype=self.model.dtype)
        np.exp(x, out=x,
               dtype=self.model.dtype)
        np.add(x, 1, out=x,
               dtype=self.model.dtype)
        np.log(x, out=y,
               dtype=self.model.dtype)
        # NOTE: Log propierty: "log(a / b) = log(a) - log(b)", and "log(1) = 0"
        np.multiply(y, -1, out=y,
                    dtype=self.model.dtype)
        return y

    def forward(self, x: np.ndarray) -> np.ndarray:
        y = self.y[:x.shape[0], :]
        log_fwd_cython(x.reshape(-1, copy=False, order="C"), y.reshape(-1, copy=False, order="C"))
        return y

    def backward(self, dy: np.ndarray) -> np.ndarray:
        dx = self.dx[:dy.shape[0], :]
        log_bwd_cython(dy.reshape(-1, copy=False, order="C"), dx.reshape(-1, copy=False, order="C"))

        return dx



from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.layer_base import LayerBase
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layers.max_pool_2d import MaxPool2D


def vgg_cyclone(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    conv_pattern = [[3, 3, 1, 64], [2, 3, 1, 128], [2, 3, 1, 256], [3, 3, 1, 512]]
    for nlayers, filter_, padding_, nfilters_ in conv_pattern:
        for layer in range(nlayers):
            _(Conv2D(nfilters=nfilters_, filter_shape=(filter_, filter_), padding=padding_, stride=1, activation=Relu))
        _(MaxPool2D(pool_shape=(2, 2), stride=2))
    _(Flatten())
    _(FC(shape=(512,), activation=Relu))
    _(FC(shape=(256,), activation=Relu))
    _(FC(shape=(128,), activation=Relu))
    _(FC(shape=(64,), activation=Relu))
    _(FC(shape=output_shape))

    return model



from pydtnn.backends.gpu.metrics.metric import MetricGPU
from pydtnn.metrics.binary_confusion_matrix import BinaryConfusionMatrix

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
import numpy as np
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore
from pydtnn.utils.constants import DTYPE2CTYPE

class BinaryConfusionMatrixGPU(BinaryConfusionMatrix[TensorGPU], MetricGPU):

    def initialize(self) -> None:
        super().initialize()
        n = self.model.batch_size
        target_classes = self.model.output_shape[0]

        self.conf_matrix = TensorGPU.create_zeros_tensor(shape=(1, target_classes, 2, 2), dtype=np.dtype(np.int32),
                                                         tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)
        self.local_cm = TensorGPU.create_zeros_tensor(shape=(n, target_classes, 2, 2), dtype=np.dtype(np.int32),
                                                      tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)
    # ----

    def __init_gpu_kernel__(self) -> Function:
        _name = "binary_confusion_matrix"
        code = """
        #define TRUE_POSITIVE  {{0,0}}
        #define TRUE_NEGATIVE  {{1,1}}
        #define FALSE_NEGATIVE {{0,1}}
        #define FALSE_POSITIVE {{1,0}}

        #define SHIFT_POINTER_CM(label, i, j, num_rows, num_columns) (((label * num_rows + i) * num_columns) + j)
        #define SHIFT_POINTER_LOCAL_CM(idx, label, i, j, num_labels, num_rows, num_columns) ((((idx * num_labels + label) * num_rows + i) * num_columns) + j)
        #define SHIFT_POINTER_Y(i, j, dim_j) (i * dim_j + j)

        __constant__ const short indexes[2][2][2] = {{
            {{TRUE_POSITIVE, TRUE_NEGATIVE}}, 
            {{FALSE_NEGATIVE, FALSE_POSITIVE}}
        }};
        
        
        __global__ void {name}({T} *y_targ, {T} *y_pred, int *cm, int *local_cm, const int num_classes, const int n)
        {{
            int label, i, j, is_pred_correct, idx;
            short index_0, index_1;
            {T} value_targ, value_pred;

            int base_idx = blockIdx.x * blockDim.x + threadIdx.x;
            const int workers = blockDim.x * gridDim.x;

            for(idx = base_idx; idx < n; idx += workers)
            {{
                for(label = 0; label < num_classes; label++)
                {{
                    value_targ = *(y_targ + SHIFT_POINTER_Y(idx, label, num_classes));
                    value_pred = *(y_pred + SHIFT_POINTER_Y(idx, label, num_classes));

                    // NOTE: y_pred[idx][label]' only possible values are 0 or 1.
                    is_pred_correct = (value_targ == value_pred);
                    index_0 = indexes[is_pred_correct][((int) value_pred)][0];
                    index_1 = indexes[is_pred_correct][((int) value_pred)][1];

                    *(local_cm + SHIFT_POINTER_LOCAL_CM(idx, label, index_0, index_1, num_classes, 2, 2)) += 1;
                }}
            }}
            // Accumulating the local values
            if (base_idx == 0)
            {{   
                for(idx = blockDim.x/2; idx > 0; idx >>= 1)
                {{
                    if(base_idx < idx)
                        for(label = 0; label < num_classes; label++)
                            for(i = 0; i < 2; i++) for(j = 0; j < 2; j++)
                                *(local_cm + SHIFT_POINTER_LOCAL_CM(base_idx, label, i, j, num_classes, 2, 2)) += *(local_cm + SHIFT_POINTER_LOCAL_CM(base_idx + idx, label, i, j, num_classes, 2, 2));
                }}
            }}
            __syncthreads();
            
            // Accumulating the local values into the output's tensor.
            if (base_idx == 0)
            {{
                for(label = 0; label < num_classes; label++)
                    for(i = 0; i < 2; i++) for(j = 0; j < 2; j++)
                        *(cm + SHIFT_POINTER_CM(label, i, j, 2, 2)) = *(local_cm + SHIFT_POINTER_LOCAL_CM(base_idx, label, i, j, num_classes, 2, 2));
            }}
        }}
        """.format(
            T = DTYPE2CTYPE[self.model.dtype],
            name = _name
        )
        module = SourceModule(code).get_function(_name)

        return module
    #---

    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU) -> np.ndarray:
        """
        For every label in target class, there is one confusion matrix like this:
                |Predicted|
        ________| T  | F  |
        Target|T| TP | FN |
              |F| FP | TN |
        """

        n = self.model.batch_size
        target_classes = self.model.output_shape[0]

        self.conf_matrix.fill(0)
        self.local_cm.fill(0)

        n = np.int32(n)
        num_classes = np.int32(target_classes)
        self.kernel(y_targ.ary, y_pred.ary, 
                    self.conf_matrix.ary, self.local_cm.ary,
                    num_classes, n,
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)

        return self.conf_matrix.ary.get()



from pydtnn.layers.layer import Layer

from pydtnn.utils.constants import Array


class Input[T: Array](Layer[T]):

    def __init__(self, shape: tuple = (1,)):
        super().__init__(shape)

    def initialize(self, prev_shape: tuple, x: T | None):
        super().initialize(prev_shape, x)



import resource
import sys
import abc

from typing import TYPE_CHECKING

from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum
from pydtnn.utils import find_component

if TYPE_CHECKING:
    from pydtnn.model import Model
    from pydtnn.layers.layer import Layer


class EventType:
    """
    EventType container
    """

    def __init__(self, name):
        self.name = name
        self._events = {}

    def __getitem__(self, item):
        try:
            description = self._events[item]
        except KeyError:
            sys.stderr.write(f"SimpleTracer warning: No event with code '{item}' "
                             f"in the '{self.name}' type of events.\n")
            return f"Unknown code {self.name}"
        return description

    def __setitem__(self, value, description):
        self._events[value] = description

    def __len__(self):
        return len(self._events)

    def items(self):
        return self._events.items()


class PostInitCaller(type):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post_init__()
        return obj


class Tracer(metaclass=PostInitCaller):
    """
    Tracer base class
    """

    def __init__(self, tracing: bool):
        self.event_types = {
            PYDTNN_MDL_EVENT: EventType("Model"),
            PYDTNN_OPS_EVENT: EventType("Operations"),
        }
        self.tracing = tracing

    def __post_init__(self):
        """
        This method will be called AFTER all the derived classes __init__ methods are completed.
        By proceeding in this way, when the derived classes enable/disable methods are called, all the attributes
        they require will already have been defined on their corresponding __init__ methods.
        """
        if self.tracing:
            self.enable_tracing()
            self.enable_print_memory_usage()
        else:
            self.disable_tracing()
            self.disable_print_memory_usage()

    def enable_tracing(self):
        """Actions that must be done if tracing is enabled"""
        setattr(self, "define_event_types", self._define_event_types)
        setattr(self, "emit_event", self._emit_event)
        setattr(self, "emit_nevent", self._emit_nevent)

    def disable_tracing(self):
        """Actions that must be done if tracing is disabled"""
        setattr(self, "define_event_types", lambda *args, **kwargs: None)
        setattr(self, "emit_event", lambda *args, **kwargs: None)
        setattr(self, "emit_nevent", lambda *args, **kwargs: None)

    def enable_print_memory_usage(self):
        """Actions that must be done if print memory usage is enabled"""
        setattr(self, "print_memory_usage", self._print_memory_usage)

    def disable_print_memory_usage(self):
        """Actions that must be done if print memory usage is disabled"""
        setattr(self, "print_memory_usage", lambda *args, **kwargs: None)

    def define_event_types(self, model: "Model"):
        """Fake method, will be replaced by lambda: None or _define_event_types()"""
        pass

    def emit_event(self, evt_type: int, evt_val: int, stream=None):
        """Fake method, will be replaced by lambda: None or _emit_event()"""
        pass

    def emit_nevent(self, evt_evt: list[int, int], evt_val: list[int, int], stream=None):
        """Fake method, will be replaced by lambda: None or _emit_nevent()"""
        pass

    def print_memory_usage(self, text: str):
        """Fake method, will be replaced by lambda: None or _print_memory_usage()"""
        pass

    def _get_layers_recursively(self, layers: list["Layer"]) -> list["Layer"]:
        all_layers = []
        for layer in layers:
            all_layers.append(layer)
            all_layers += self._get_layers_recursively(layer.children)
        return all_layers

    def _define_event_types(self, model: "Model"):
        """This method will be called only if tracing is enabled"""
        mdl_event = self.event_types[PYDTNN_MDL_EVENT]
        ops_event = self.event_types[PYDTNN_OPS_EVENT]
        mdl_event[0] = "End"
        ops_event[0] = "End"
        mdl_constants = [(event._name_, event._value_) for event in PYDTNN_MDL_EVENT_enum]
        ops_constants = [(event._name_, event._value_) for event in PYDTNN_OPS_EVENT_enum]
        for layer in model.get_all_layers():
            for (name, val) in mdl_constants:
                mdl_event[layer.id * PYDTNN_MDL_EVENTS + val] = f"{layer.name_with_id}_{name[11:].lower()}"
            for (name, val) in ops_constants:
                ops_event[
                    layer.id * PYDTNN_OPS_EVENTS + val] = f"{layer.id:03}_{layer.name}_{name[11:].lower()}"

    @abc.abstractmethod
    def _emit_event(self, evt_type: int, evt_val: int, stream=None):
        """This method will be called only if tracing is enabled"""
        pass

    @abc.abstractmethod
    def _emit_nevent(self, evt_evt: list[int], evt_val: list[int], stream=None):
        """This method will be called only if tracing is enabled"""
        pass

    @staticmethod
    def _print_memory_usage(text=""):
        """This method will be called only if print memory usage is enabled"""
        u = resource.getrusage(resource.RUSAGE_SELF)
        if text != "":
            text = f" {text}:"
        print(f">>>{text} user time={u[0]:.2f}, sys time={u[1]:.2f}, mem={u[2] / 1024:.2f} MiB")

    def set_stream(self, stream):
        pass


def select(name: str) -> type[Tracer]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.layers.abstract.block_layer import AbstractBlockLayer


class AbstractBlockLayerGPU(AbstractBlockLayer[TensorGPU], LayerGPU):
    pass



import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.categorical_accuracy import CategoricalAccuracy


class CategoricalAccuracyCPU(CategoricalAccuracy[np.ndarray], MetricCPU):

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        b = y_targ.shape[0]
        # return np.sum(y_targ[np.arange(b), np.argmax(y_pred, axis=1)]) * 100 / b
        y = y_targ[np.arange(b), np.argmax(y_pred, axis=1)]
        y = np.sum(y, dtype=self.model.dtype)
        y *= 100 / b
        return y



from typing import TYPE_CHECKING
import time

import numpy as np

from pydtnn.schedulers.scheduler_with_loss_or_metric import SchedulerWithLossOrMetric

if TYPE_CHECKING:
    from pydtnn.model import Model


class EarlyStopping(SchedulerWithLossOrMetric):
    """
    EarlyStopping LRScheduler
    """

    def __init__(self, loss_or_metric="", patience=10, minimize=True, verbose=True):
        super().__init__(loss_or_metric, verbose)
        self.patience = patience
        self.minimize = minimize
        self.best_epoch: int = 0
        self.best_loss: float = np.inf * {True: -1, False: 1}[not self.minimize]
        self.best_weights_filename: str | None = None
        self.time: str = time.strftime("%Y%m%d")

    def on_epoch_end(self, train_loss: np.ndarray[float], val_loss: np.ndarray[float]) -> None:
        idx = self._get_idx()
        self.epoch_count += 1
        loss = val_loss if self.is_val_metric else train_loss
        if (not self.minimize and loss[idx] > self.best_loss) or \
                (self.minimize and loss[idx] < self.best_loss):
            self.best_loss = loss[idx]
            self.best_epoch = self.epoch_count
            # Save weights + bias
            if not self.best_weights_filename:
                self.best_weights_filename = "./model-{}-weights-rank_{}-{}.npz" \
                    .format(self.model.model_name, self.model.comm_rank, self.time)
            self.model.store_weights_and_bias(self.best_weights_filename, compress=False)
        elif (self.epoch_count - self.best_epoch) >= self.patience:
            self.stop_training = True
            # Restore weights + bias
            self.model.load_weights_and_bias(self.best_weights_filename)
            self.log(f"Metric '{self.loss_or_metric}' did not improve for {self.patience} epochs, stop training.")

    @classmethod
    def from_model(cls, model: "Model") -> "EarlyStopping":
        return EarlyStopping(model.early_stopping_metric,
                             model.early_stopping_patience,
                             model.early_stopping_minimize)



"""
Tracer events
"""
from enum import IntEnum, auto

# ========================== #
# ========= COMMON ========= #
# ========================== #
PYDTNN_EVENT_FINISHED = 0
# ========================== #


# ========================== #
# ==== PYDTNN_MDL_EVENT ==== #
# ========================== #
class PYDTNN_MDL_EVENT_enum(IntEnum):
    FORWARD = auto()  # Originally: 1
    BACKWARD = auto()  # Originally: 2
    ALLREDUCE_DW = auto()  # Originally: 3
    WAIT_DW = auto()  # Originally: 4
    UPDATE_DW = auto()  # Originally: 5



PYDTNN_MDL_EVENT = 60000001
PYDTNN_MDL_EVENTS = len(PYDTNN_MDL_EVENT_enum)
# ========================== #

# ========================== #
# ==== PYDTNN_OPS_EVENT ==== #
# ========================== #


class PYDTNN_OPS_EVENT_enum(IntEnum):
    OPS_ALLREDUCE_DW = auto()  # Orginally: 1
    BACKWARD_CONVGEMM = auto()  # Orginally: 2
    BACKWARD_CUBLAS_MATMUL_DW = auto()  # Orginally: 3
    BACKWARD_CUBLAS_MATMUL_DX = auto()  # Orginally: 4
    BACKWARD_CUBLAS_MATVEC_DB = auto()  # Orginally: 5
    BACKWARD_CUDNN_DB = auto()  # Orginally: 6
    BACKWARD_CUDNN_DW = auto()  # Orginally: 7
    BACKWARD_CUDNN_DX = auto()  # Orginally: 8
    BACKWARD_DECONV_GEMM = auto()  # Orginally: 9
    BACKWARD_ELTW_SUM = auto()  # Orginally: 10
    BACKWARD_IM2COL = auto()  # Orginally: 11
    BACKWARD_RESHAPE_DW = auto()  # Orginally: 12
    BACKWARD_RESHAPE_DX = auto()  # Orginally: 13
    BACKWARD_SPLIT = auto()  # Orginally: 14
    BACKWARD_SUM_BIASES = auto()  # Orginally: 15
    BACKWARD_TRANSPOSE_DY = auto()  # Orginally: 16
    BACKWARD_TRANSPOSE_W = auto()  # Orginally: 17
    BACKWARD_ADP_AVG_POOL = auto()  # Now: 18
    COMP_DW_MATMUL = auto()  # Orginally: 18
    COMP_DX_COL2IM = auto()  # Orginally: 19
    COMP_DX_MATMUL = auto()  # Orginally: 20
    FORWARD_DEPTHWISE_CONV = auto()  # Orginally: 21
    FORWARD_POINTWISE_CONV = auto()  # Orginally: 22
    FORWARD_CONCAT = auto()  # Orginally: 23
    FORWARD_CONVGEMM = auto()  # Orginally: 24
    FORWARD_CONVWINOGRAD = auto()  # Orginally: 25
    FORWARD_CONVDIRECT = auto()  # Orginally: 26
    FORWARD_CUBLAS_MATMUL = auto()  # Orginally: 27
    FORWARD_CUDNN = auto()  # Orginally: 28
    FORWARD_CUDNN_SUM_BIASES = auto()  # Orginally: 29
    FORWARD_ELTW_SUM = auto()  # Orginally: 30
    FORWARD_IM2COL = auto()  # Orginally: 31
    FORWARD_MATMUL = auto()  # Orginally: 32
    FORWARD_REPLICATE = auto()  # Orginally: 33
    FORWARD_RESHAPE_W = auto()  # Orginally: 34
    FORWARD_RESHAPE_Y = auto()  # Orginally: 35
    FORWARD_SUM_BIASES = auto()  # Orginally: 36
    FORWARD_TRANSPOSE_Y = auto()  # Orginally: 37
    FORWARD_MHA_FC_QKV = auto()  # Orginally: 38
    FORWARD_MHA_MATMUL_QK = auto()  # Orginally: 39
    FORWARD_MHA_SCALARDK = auto()  # Orginally: 40
    FORWARD_MHA_MATMUL_SMV = auto()  # Orginally: 41
    FORWARD_MHA_FC_O = auto()  # Orginally: 42
    BACKWARD_MHA_FC_QKV = auto()  # Orginally: 43
    BACKWARD_MHA_MATMUL_QK = auto()  # Orginally: 44
    BACKWARD_MHA_SCALARDK = auto()  # Orginally: 45
    BACKWARD_MHA_MATMUL_SMV = auto()  # Orginally: 46
    BACKWARD_MHA_FC_O = auto()  # Orginally: 47
    FORWARD_MHA = auto()  # Orginally: 48
    FORWARD_FEEDFORWARD = auto()  # Orginally: 49
    BACKWARD_MHA = auto()  # Orginally: 50
    BACKWARD_FEEDFORWARD = auto()   # Orginally: 51
    FORWARD_ADP_AVG_POOL = auto()  # Orginally: 39


PYDTNN_OPS_EVENT = 60000002
PYDTNN_OPS_EVENTS = len(PYDTNN_OPS_EVENT_enum)
# ========================== #



"""TenSEAL encryption"""

# NOTE: Dataclasses with slots can not use zero-arg super() (gh-90562)
# FIXME: Serialization preformance

import sys
import pickle
import copyreg
import dataclasses
from dataclasses import dataclass

import numpy as np

from pydtnn.libs import libcrypt

# Make sure global package is not confused with current package
_pkg = sys.path.pop(0)
try:
    import tenseal
    from tenseal import sealapi
    from tenseal.tensors import CKKSVector
    from tenseal.enc_context import Context as SealContext
finally:
    sys.path.insert(0, _pkg)


__all__ = (
    "Context",
)


SECURITY_LEVEL = {
    0: sealapi.SEC_LEVEL_TYPE.NONE,
    128: sealapi.SEC_LEVEL_TYPE.TC128,
    192: sealapi.SEC_LEVEL_TYPE.TC192,
    256: sealapi.SEC_LEVEL_TYPE.TC256
}


@dataclass(eq=False, order=False, slots=True, frozen=True)
class Ciphertext[P: np.number](libcrypt.Ciphertext[CKKSVector, P]):
    """TenSEAL ciphertext"""
    _context: bytes = dataclasses.field(repr=False)

    def _new(self, /, *args, **kwds):
        """Create new operable ciphertext"""
        return super(Ciphertext, self)._new(_context=self._context, *args, **kwds)

    def _operable(self, other) -> None:
        """Ensure ciphertext is operable"""
        super(Ciphertext, self)._operable(other)

        if other._context != self._context:
            raise TypeError(f"Different underlying context ({hash(other._context)} != {hash(self._context)})")

        # Synchronize contexts
        context = self._get_context() or other._get_context() or self._load_context()
        self._link_context(context)
        other._link_context(context)

    def _link_context(self, context: SealContext):
        """Link all chunks to context"""
        for chunk in self._chunks:
            chunk.link_context(context)

    def _load_context(self) -> SealContext:
        """Load stored context"""
        return pickle.loads(self._context)

    def _get_context(self) -> SealContext | None:
        """Get loaded context"""
        if not self._chunks:
            return None
        chunk = self._chunks[0]
        try:
            return chunk.context()
        except ValueError:
            return None


class Context(libcrypt.Context[CKKSVector]):
    """TenSEAL context"""
    _cls = Ciphertext

    def __init__(self, poly_degree: int = 13, global_scale: int = 40, security_level: int = 128) -> None:
        """Inizialize context"""
        super().__init__(poly_degree, global_scale, security_level)

        poly_degree = 2 ** self._poly_degree
        level = SECURITY_LEVEL[self._security_level]

        # Context
        modulus = [
            m.bit_count()
            for m in sealapi.CoeffModulus.BFVDefault(poly_degree, level)
        ]
        self._private_context = tenseal.context(
            scheme=tenseal.SCHEME_TYPE.CKKS,
            poly_modulus_degree=poly_degree,
            coeff_mod_bit_sizes=modulus
        )

        # Keys
        self._private_context.global_scale = 2 ** self._global_scale
        self._private_context.generate_galois_keys()
        self._private_context.generate_relin_keys()

        # Public
        self._public_context = self._private_context.copy()
        self._public_context.make_context_public()

        self._context = pickle.dumps(self._public_context)

    def _new(self, /, *args, **kwds) -> libcrypt.Ciphertext:
        """Create new operable ciphertext"""
        return super()._new(_context=self._context, *args, **kwds)

    def _encrypt_chunk(self, chunk: list) -> CKKSVector:
        """Encode list to ciphertext"""
        return tenseal.ckks_vector(self._public_context, chunk)

    def _decrypt_chunk(self, chunk: CKKSVector) -> list:
        """Decode cypertext to list"""
        chunk.link_context(self._private_context)
        return chunk.decrypt()


# Pickle support
def context_reducer(context: SealContext):
    """TenSEAL context pickle reducer"""
    cls = context.load
    args = (context.serialize(save_secret_key=True),)
    return (cls, args)


def ckks_vector_reducer(vector: CKKSVector):
    """TenSEAL CKKS vector pickle reducer"""
    cls = vector.lazy_load
    args = (vector.serialize(),)
    return (cls, args)


copyreg.pickle(SealContext, context_reducer)
copyreg.pickle(CKKSVector, ckks_vector_reducer)  # type: ignore (wrong inferred typing)



import numpy as np
import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.compiler import SourceModule  #type: ignore
from pycuda.driver import Function  #type: ignore

from pydtnn.losses.categorical_cross_entropy import CategoricalCrossEntropy
from pydtnn.backends.gpu.losses.loss import LossGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import DTYPE2CTYPE


class CategoricalCrossEntropyGPU(LossGPU, CategoricalCrossEntropy[TensorGPU]):

    def __init_gpu_kernel__(self) -> Function:
        _name = "categorical_cross_entropy"
        code ="""
        __global__ void {name}({T} *y_targ, {T} *y_pred, {T} *res,
                               {T} *dx, int b, int n, float eps)
        {{
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            if (idx < b) 
            {{
                int i = 0, max = 0;
                {T} max_value = y_targ[idx * n];
                dx[idx * n] = y_targ[idx * n];
                for ( i = 1; i < n; i++ ) 
                {{
                    dx[idx * n + i] = y_targ[idx * n + i];
                    if ( y_targ[idx * n + i] > max_value ) 
                    {{
                        max = i;
                        max_value = y_targ[idx * n + i];
                    }}
                }}
                
                {T} pred = y_pred[idx * n + max];
                if ( pred < eps )          pred = eps;
                else if ( pred > (1-eps) ) pred = (1-eps);

                res[idx] = logf(pred);
                dx[idx * n + max] /= -(pred * b);
            }}
            return;
        }}
        """.format(
            T=DTYPE2CTYPE[self.model.dtype],
            name=_name
        )

        module = SourceModule(code).get_function(_name)

        return module

    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU, batch_size: int) -> tuple[float, TensorGPU]:
        self.kernel(y_targ.ary, y_pred.ary, self.loss, self.dx.ary,
                    np.int32(batch_size), np.int32(self.shape[1]), np.float32(self.eps),
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)
        loss: float = -gpuarray.sum(self.loss[:batch_size]).get() / batch_size
        return loss, self.dx



import numpy as np

from pydtnn.activations.arctanh import Arctanh
from pydtnn.backends.gpu.activations.activation import ActivationGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU

import pycuda.gpuarray as gpuarray # type: ignore
from pycuda.elementwise import ElementwiseKernel # type: ignore
from pydtnn.utils.constants import ArrayShape, DTYPE2CTYPE


class ArctanhGPU(Arctanh[TensorGPU], ActivationGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.atanh: ElementwiseKernel = None
        self.datanh: ElementwiseKernel = None

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        self.atanh = ElementwiseKernel(
            "{T} *in, {T} *out".format(T=DTYPE2CTYPE[self.model.dtype]),
            "out[i] = {func}(in[i]);" .format(func={np.float32: "atanhf", np.float64: "atanh"}[self.model.dtype]),
            "atanh")

        self.datanh = ElementwiseKernel(
            "{T} *in, {T} *out".format("T", DTYPE2CTYPE[self.model.dtype]),
            "out[i] = 1.0 / (1.0 + {func}(in[i], 2));".format(func={np.float32: "powf", np.float64: "pow"}[self.model.dtype]),
            "datanh")

        # Activations y
        y_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

    def forward(self, x: TensorGPU) -> TensorGPU:
        self.atanh(x.ary, self.y.ary, stream=self.model.stream)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        # Compute dx
        self.datanh(dy.ary, self.dx.ary, stream=self.model.stream)
        return self.dx



import numpy as np

from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.dropout import Dropout
from pydtnn.model import Model
from pydtnn.utils import random


class DropoutCPU(Dropout[np.ndarray], LayerCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mask: np.ndarray = None  # type: ignore (It will be initalized later.)

    def forward(self, x: np.ndarray) -> np.ndarray:

        match self.model.mode:
            case Model.Mode.TRAIN:
                # NOTE: Remember, it's necessary a new random mask every training's forward call.
                # self.mask = random.binomial(1, (1 - self.rate), size=self.shape).astype(self.model.dtype) / (1 - self.rate)
                self.mask = np.asarray(random.binomial(n=1, p=(1 - self.rate), size=self.shape), dtype=self.model.dtype, order="C", copy=None)
                np.divide(self.mask, (1 - self.rate), out=self.mask, dtype=self.model.dtype)
                np.multiply(x, self.mask, out=x, order="C", dtype=self.model.dtype)
            case Model.Mode.EVALUATE:
                pass  # Just returns x.
            case _:
                raise RuntimeError(f"Unexpected model mode \'{self.model.mode}\'.")
        return x
    # ----

    def backward(self, dy: np.ndarray) -> np.ndarray:
        np.multiply(dy, self.mask, out=dy, dtype=self.model.dtype, order="C")
        return dy



import importlib
from typing import TYPE_CHECKING, Optional

from pydtnn.layer_base import FusedLayerMixIn
if TYPE_CHECKING:
    from pydtnn.activations.activation import Activation
from pydtnn.backends import BackendType
from pydtnn.layers.layer import Layer
from pydtnn.utils.initializers import InitializerFunc, glorot_uniform, zeros
import numpy as np
from enum import StrEnum, auto
from pydtnn.utils.constants import Array, ArrayShape, Parameters


class Conv2D[T: Array](Layer[T]):

    class Grouping(StrEnum):
        DEPTHWISE = auto()
        POINTWISE = auto()
        STANDARD = auto()
    # -------

    class Variant(StrEnum):
        BEST_OF = auto()
        I2C = auto()
        #NOTE: The following values are not set by auto due it's necessary that have that value.
        #TODO: Check how to change this (BestOf and Fusion layers)
        GEMM = "cg"
        WINOGRAD = "cw"
        DIRECT = "cd0"
    # -----

    def _get_backend_cls(self) -> None:
        if isinstance(self, FusedLayerMixIn):
            return super()._get_backend_cls()

        cls = self.__class__
        backend = self.model._backend
        module_name = cls.__module__.split(".", 1)[1]

        if backend is BackendType.CPU and self.grouping is self.Grouping.STANDARD:
            variant = self.model.conv_variant._name_.lower()
        else:
            variant = self.grouping.lower()

        backend_module_name = f"pydtnn.backends.{backend}.{module_name}_variants.{variant}"
        try:
            backend_module = importlib.import_module(backend_module_name)
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError("Check the package were the variants are located; it must have the same name as \"{module_name}_variants\" above.") from e
        cls_name = f"{cls.__name__}{variant.title()}{backend.upper()}"
        cls = getattr(backend_module, cls_name)
        return cls

    def __init__(self, nfilters: int = 1,
                 filter_shape: tuple[int, int] | int = (3, 3),
                 grouping: Grouping = Grouping.STANDARD,
                 padding: tuple[int, int] | int = 0,
                 stride: tuple[int, int] | int = 1,
                 dilation: tuple[int, int] | int = 1,
                 activation: Optional[type["Activation"]] = None,
                 use_bias=True,
                 weights_initializer: InitializerFunc = glorot_uniform,
                 biases_initializer: InitializerFunc = zeros):

        super().__init__()
        self.co = nfilters
        self.filter_shape = (filter_shape, filter_shape) if isinstance(filter_shape, int) else filter_shape
        self.grouping = Conv2D.Grouping(grouping.lower())
        self.padding = padding
        self.stride = stride
        self.dilation = dilation
        self.vpadding, self.hpadding = (padding, padding) if isinstance(padding, int) else padding
        self.vstride, self.hstride = (stride, stride) if isinstance(stride, int) else stride
        self.vdilation, self.hdilation = (dilation, dilation) if isinstance(dilation, int) else dilation
        self.act = activation
        self.use_bias = use_bias
        self.weights_initializer: InitializerFunc = weights_initializer
        self.biases_initializer: InitializerFunc = biases_initializer
        self.grad_vars = {Parameters.WEIGHTS: Parameters.DW}
        if self.use_bias:
            self.grad_vars[Parameters.BIASES] = Parameters.DB
        self.debug = False
        # The following attributes will be initialized later
        self.ci = self.hi = self.wi = self.kh = self.kw = self.ho = self.wo = 0
        self.weights_shape: ArrayShape = None  # type: ignore
        # @warning: do not do this (affects the gpu version) self.forward = self.backward = None
    # ---

    def _initializing_special_parameters(self):
        # NOTE: This method's objective is to define and change the value of some parameters defined before that are needed later in the initialization process,
        #   for example: "self.weights_shape" and, in non-standard cases, "self.co".
        pass
    # ---

    def initialize(self, prev_shape: ArrayShape, x: T | None):
        super().initialize(prev_shape, x)
        self.ci, self.hi, self.wi = self.model.decode_shape(prev_shape)
        self.kh, self.kw = self.filter_shape
        self._initializing_special_parameters()

        self.ho = (self.hi + 2 * self.vpadding - self.vdilation * (self.kh - 1) - 1) // self.vstride + 1
        self.wo = (self.wi + 2 * self.hpadding - self.hdilation * (self.kw - 1) - 1) // self.hstride + 1
        self.shape = self.model.encode_shape((self.co, self.ho, self.wo))

        # NOTE: self.weights_shape must be defined in "self._initializing_special_parameters"
        self.nparams = int(np.prod(self.weights_shape) + (self.co if self.use_bias else 0))
    # --

    def _show_props(self) -> dict:
        props = super()._show_props()

        props["padding"] = (self.vpadding, self.hpadding)
        props["stride"] = (self.vstride, self.hstride)
        props["dilation"] = (self.vdilation, self.hdilation)

        return props



from warnings import warn

# Layers
try:
    from pydtnn.tests.layer_pytorch import LayerPyTorchTestCase
except Exception:
    warn("PyTorch not available, skiping tests!")

# Fused
from pydtnn.tests.conv_2d_group import Conv2DGroupTestCase
from pydtnn.tests.conv_2d_relu import Conv2DReluTestCase
from pydtnn.tests.conv_2d_batch_normalization import Conv2DBatchNormalizationTestCase
from pydtnn.tests.conv_2d_batch_normalization_relu import Conv2DBatchNormalizationReluTestCase
from pydtnn.tests.batch_normalization_relu import BatchNormalizationReluTestCase

# Models
from pydtnn.tests.model_dtype import ModelDTypeTestCase
from pydtnn.tests.model_tensor import ModelTensorTestCase
try:
    from pydtnn.tests.model_gpu import ModelGpuTestCase
except Exception:
    warn("GPU not available, skiping tests!")

# Libraries
try:
    from pydtnn.tests.conv_gemm import ConvGemmTestCase
    from pydtnn.tests.conv_2d_conv_gemm import Conv2DConvGemmTestCase
    from pydtnn.tests.model_conv_gemm import ModelConvGemmTestCase
except Exception:
    warn("ConvGemm not available, skiping tests!")

try:
    from pydtnn.tests.conv_winograd import ConvWinogradTestCase
except Exception:
    warn("ConvWinograd not available, skiping tests!")

try:
    from pydtnn.tests.conv_direct import ConvDirectTestCase
except Exception:
    warn("ConvDirect not available, skiping tests!")



import numpy as np
from pydtnn.layers.layer import Layer

from pydtnn.utils.constants import Array
from pydtnn.utils.constants import ArrayShape


class Flatten[T: Array](Layer[T]):

    def initialize(self, prev_shape: ArrayShape, x: T | None):
        super().initialize(prev_shape, x)
        self.shape = (int(np.prod(prev_shape)),)



import unittest

from pydtnn.layers.layer import LayerError
from pydtnn.model import Model

from pydtnn.tests.abstract.common import Params
from pydtnn.tests.abstract.model_common import ModelCommonTestCase
from pydtnn.utils.tensor import TensorFormat


class ModelConvGemmTestCase(ModelCommonTestCase):
    """
    Tests that two models with different parameters lead to the same results
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global ModelCommonTestCase
    del ModelCommonTestCase

    # Compares results between an XX model {self.model1_desc} and other {self.model1_desc}
    model1_desc = "using Im2Col+MM"
    model2_desc = "using ConvGemm"

    def get_model2(self, model_name: str, overwrite_params: dict | None = None) -> Model:
        # CPU model with convGemm
        params = Params()
        # Begin of params configuration
        params.model_name = model_name
        params.conv_variant = "gemm"
        params.tensor_format = TensorFormat.NHWC.upper()
        # End of params configuration
        params_dict = vars(params)
        if overwrite_params is not None:
            params_dict.update(overwrite_params)
        try:
            model2 = Model(**params_dict)
        except LayerError as exc:
            raise unittest.SkipTest(f"Model {model_name} incompatible with {params_dict['dataset_name']}") from exc
        return model2



from pydtnn.backends.cpu.layers.abstract.conv_2d_standard import Conv2DStandardCPU
from pydtnn.backends.cpu.layers.batch_normalization import BatchNormalizationCPU
from pydtnn.layers.conv_2d_batch_normalization import Conv2DBatchNormalization
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.utils.constants import ArrayShape, Parameters

import numpy as np


class Conv2DBatchNormalizationCPU(Conv2DBatchNormalization[np.ndarray], Conv2DStandardCPU):

    @property
    def _ary_prop(self) -> set[str]:
        return {Parameters.RUNNING_MEAN, 
                Parameters.RUNNING_VAR, 
                *super()._ary_prop}


    # NOTE: The "__init__" method is being made (more or less) in Model (in _apply_layer_fusion) and in FusedLayerMixIn.
    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None) -> None:
        super().initialize(prev_shape, x)

        self.inv_std = BatchNormalizationCPU.get_inv_std(self.running_var, self.epsilon, self.model.dtype)

        self.forward = {"_forward_cw_nchw": self._forward_nchw_cw,
                        "_forward_cg_nchw": self._forward_nchw_cg,
                        "_forward_cg_nhwc": self._forward_nhwc_cg}[self.forward.__name__]
        self.backward = self._backward

    def _forward_nchw_cw(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convWinograd + BatchNorm + """
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        y: np.ndarray = self.cw.conv_winograd_nchw(self.weights, x, self.biases,
                                                vpadding=self.vpadding, hpadding=self.hpadding,
                                                vstride=self.vstride, hstride=self.hstride,
                                                vdilation=self.vdilation, hdilation=self.hdilation,
                                                relu=False, bn=True,
                                                running_mean=self.running_mean,
                                                inv_std=self.inv_std, gamma=self.gamma, beta=self.beta)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _forward_nchw_cg(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convGemm + BatchNorm"""
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        res: np.ndarray = self.cg.conv_gemm_nchw(self.weights, x,
                                              vpadding=self.vpadding, hpadding=self.hpadding,
                                              vstride=self.vstride, hstride=self.hstride,
                                              vdilation=self.vdilation, hdilation=self.hdilation,
                                              biases=self.biases, bn_running_mean=self.running_mean,
                                              bn_inv_std=self.inv_std, bn_gamma=self.gamma, bn_beta=self.beta, relu=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(res, dtype=self.model.dtype, order='C', copy=None)

    def _forward_nhwc_cg(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convGemm + BatchNorm"""
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        res: np.ndarray = self.cg.conv_gemm_nhwc(self.weights, x,
                                              vpadding=self.vpadding, hpadding=self.hpadding,
                                              vstride=self.vstride, hstride=self.hstride,
                                              vdilation=self.vdilation, hdilation=self.hdilation,
                                              biases=self.biases, bn_running_mean=self.running_mean,
                                              bn_inv_std=self.inv_std, bn_gamma=self.gamma, bn_beta=self.beta, relu=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(res, dtype=self.model.dtype, order='C', copy=None)

    def _backward(self, dy: np.ndarray) -> np.ndarray:
        raise NotImplementedError("Use a real backward variant!")



from pydtnn.optimizers.optimizer import Optimizer
import numpy as np


class OptimizerCPU(Optimizer[np.ndarray]):
    """
    Extends an Optimizer class with the attributes and methods required by CPU Optimizers.
    """

    def are_all_zeros(self, ndarray: np.ndarray) -> bool:
        return not ndarray.any()



import ctypes

import pycuda.driver as drv  # type: ignore


# The below code will allocate the maximum used memory, which will be shared
# among all layers. This code saves having a memory allocation per layer.
ws_size = 1
ws = drv.mem_alloc(ws_size) if ws_size > 0 else 0
ws_ptr = ctypes.c_void_p(int(ws))


def checkConvolutionMemory(size):
    global ws_size
    global ws
    global ws_ptr
    # if a layer requires more memory than the allocated
    # we re-allocated that size
    if size.value > ws_size:
        ws_size = size.value
        assert not isinstance(ws, int), f"\"ws\" must not be an \"int\" here ({type(ws)=} || {ws=})."
        ws.free()
        ws = drv.mem_alloc(ws_size) if ws_size > 0 else 0
        ws_ptr = ctypes.c_void_p(int(ws))


def getConvolutionWorkspaceSize():
    return ws_size


def getConvolutionWorkspacePtr():
    return ws_ptr



import numpy as np

from pydtnn.backends.cpu.layers.abstract.conv_2d_standard import Conv2DStandardCPU
from pydtnn.libs.libconvgemm import ConvGemm
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.utils.constants import ArrayShape
from pydtnn.utils.tensor import TensorFormat


class Conv2DGemmCPU(Conv2DStandardCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # convGemm related attributes (will be initialized in initialize())
        self.cg = None # type: ignore

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)
        # ConvGemm parameters
        self.cg: ConvGemm = ConvGemm(dtype=self.model.dtype, debug=self.debug, parent_layer=self)

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.forward = self._forward_cg_nchw
                self.backward = self._backward_cg_nchw
            case TensorFormat.NHWC:
                self.forward = self._forward_cg_nhwc
                self.backward = self._backward_cg_nhwc
            case _:
                raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
        # ---
    # ----

    def _forward_cg_nhwc(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convGemm library"""

        self.cg_x = x

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        y: np.ndarray = self.cg.conv_gemm_nhwc(self.weights, x,
                                               vpadding=self.vpadding, hpadding=self.hpadding,
                                               vstride=self.vstride, hstride=self.hstride,
                                               vdilation=self.vdilation, hdilation=self.hdilation,
                                               biases=self.biases)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return y

    def _forward_cg_nchw(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convGemm library"""

        self.cg_x = x

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        res = self.cg.conv_gemm_nchw(self.weights, x,
                                     vpadding=self.vpadding, hpadding=self.hpadding,
                                     vstride=self.vstride, hstride=self.hstride,
                                     vdilation=self.vdilation, hdilation=self.hdilation,
                                     biases=self.biases)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return res

    def _backward_cg_nhwc(self, dy: np.ndarray) -> np.ndarray:
        """Version of the backward function that uses the convGemm library"""
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CONVGEMM)
        res: np.ndarray = np.zeros(self.weights.shape, dtype=dy.dtype)
        self.cg.conv_gemm_nhwc(dy, self.cg_x, out=res,
                               vpadding=self.vpadding, hpadding=self.hpadding,
                               vstride=self.vstride, hstride=self.hstride,
                               vdilation=self.vdilation, hdilation=self.hdilation,
                               trans=True)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        self.dw = res

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
        if self.use_bias:
            self.db: np.ndarray = np.sum(dy, axis=(0, 1, 2))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                     self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_DECONV_GEMM)
        dx: np.ndarray = np.zeros((dy.shape[0], self.hi, self.wi, self.ci), dtype=dy.dtype)
        self.cg.deconv_gemm_nhwc(self.weights, dy, dx,
                                 vpadding=self.vpadding, hpadding=self.hpadding,
                                 vstride=self.vstride, hstride=self.hstride,
                                 vdilation=self.vdilation, hdilation=self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return dx

    def _backward_cg_nchw(self, dy: np.ndarray) -> np.ndarray:
        """Version of the backward function that uses the convGemm library"""
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CONVGEMM)
        res = np.zeros(self.weights.shape, dtype=dy.dtype)
        # NOTE: conv_gemm_nchw, in this context seems that is being used as a matrix multiplication instead of a convolution.
        self.cg.conv_gemm_nchw(dy, self.cg_x, out=res,
                               vpadding=self.vpadding, hpadding=self.hpadding,
                               vstride=self.vstride, hstride=self.hstride,
                               vdilation=self.vdilation, hdilation=self.hdilation,
                               trans=True)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        self.dw = res

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
        if self.use_bias:
            self.db = np.sum(dy, axis=(0, 2, 3))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                     self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_DECONV_GEMM)
        dx = np.zeros((dy.shape[0], self.ci, self.hi, self.wi), dtype=dy.dtype)
        self.cg.deconv_gemm_nchw(self.weights, dy, dx,
                                 vpadding=self.vpadding, hpadding=self.hpadding,
                                 vstride=self.vstride, hstride=self.hstride,
                                 vdilation=self.vdilation, hdilation=self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return dx



import numpy as np

from pydtnn.backends.cpu.layers.abstract.block_layer import AbstractBlockLayerCPU
from pydtnn.layers.multi_head_attention import MultiHeadAttention
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.feed_forward import FeedForward
from pydtnn.layers.layer_normalization import LayerNormalization
from pydtnn.layers.decoder import Decoder

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_OPS_EVENT_enum


class DecoderCPU(Decoder[np.ndarray], AbstractBlockLayerCPU):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.multiheadattention = MultiHeadAttention(embedl=self.embedl, d_k=self.d_k, heads=self.heads, dropout_rate=self.dropout_rate)
        # self.dropout_1 = Dropout(rate=self.dropout_rate)
        self.layernormalization_1 = LayerNormalization(axis=(1, 2))
        self.multiheadattention_enc = MultiHeadAttention(embedl=self.embedl, d_k=self.d_k, heads=self.heads, dropout_rate=self.dropout_rate)
        # self.dropout_enc = Dropout(rate=self.dropout_rate)
        self.layernormalization_enc = LayerNormalization(axis=(1, 2))
        self.feedforward = FeedForward(shape=(self.embedl,), d_ff=self.d_ff, dropout_rate=self.dropout_rate)
        self.dropout_2 = Dropout(rate=self.dropout_rate)
        self.layernormalization_2 = LayerNormalization(axis=(1, 2))
        # self.paths = [[self.multiheadattention, self.dropout_1, self.layernormalization_1, self.multiheadattention_enc,
        #                self.dropout_enc, self.layernormalization_enc, self.feedforward, self.dropout_2, self.layernormalization_2]]
        self.paths = [[self.multiheadattention, self.layernormalization_1, self.multiheadattention_enc,
                       self.layernormalization_enc, self.feedforward, self.dropout_2, self.layernormalization_2]]

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        x_dec, x_enc, mask_dec = x
        x_dec_shape, x_enc_shape, mask_dec_shape = prev_shape
        mha_shape = (x_dec_shape, mask_dec_shape)

        # Initialize all sublayers
        for layer in self.children:
            layer.init_backend_from_model(self.model)

        self.multiheadattention.initialize(prev_shape=mha_shape, x=(x_dec, x_dec, x_dec, mask_dec))
        self.layernormalization_1.initialize(prev_shape=self.shape, x=self.multiheadattention.y)
        self.multiheadattention_enc.initialize(prev_shape=mha_shape, x=(x_dec, x_enc, x_enc, mask_dec))
        # self.dropout_enc.initialize(prev_shape=prev_shape, x=self.multiheadattention_enc.y)
        self.layernormalization_enc.initialize(prev_shape=self.shape, x=self.multiheadattention_enc.y)

        # x_aux = self.flatten(self.layernormalization_enc.y)
        self.layernormalization_enc_y_flatten = None  # x_aux.copy()

        self.feedforward.initialize(prev_shape=self.layernormalization_enc.shape, x=self.layernormalization_enc_y_flatten)
        # x_aux = self.unflatten(self.feedforward.dx)
        self.feedforward_dx_unflatten = None  # x_aux.copy()

        self.dropout_2.initialize(prev_shape=self.feedforward.shape, x=self.feedforward_y_unflatten)

        self.layernormalization_2.initialize(prev_shape=prev_shape, x=self.dropout_2.y)

        self.y = self.layernormalization_2.y

        for layer in self.children:
            self.fwd_time += layer.fwd_time
            self.bwd_time += layer.bwd_time
            self.nparams += layer.nparams

    def initialize_block_layer(self):
        pass

    def flatten(self, x):
        last_dim = x.ary.shape[-1]
        return x.reshape((int(np.prod(self.first_dims)), last_dim))

    def unflatten(self, x):
        last_dim = x.ary.shape[-1]
        return x.reshape((*self.first_dims, last_dim))

    def forward(self, x, x_enc, mask=None):
        # Self Attention
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MHA)
        x_1 = self.multiheadattention.forward(x, x, x, mask)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
        # x_1 = self.dropout_1.forward(x_1)
        x_1 += x
        x_1 = self.layernormalization_1.forward(x_1)
        # Encoder-Decoder Attention
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MHA)
        x_2 = self.multiheadattention_enc.forward(x_1, x_enc, x_enc, mask)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
        # x_2 = self.dropout_enc.forward(x_2)
        x_2 += x_1
        x_2 = self.layernormalization_enc.forward(x_2)
        # Feed Forward
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_FEEDFORWARD)
        x_3 = self.feedforward.forward(x_2)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
        x_3 = self.dropout_2.forward(x_3)
        x_3 += x_2
        x_3 = self.layernormalization_2.forward(x_3)
        return x_3

    def backward(self, dy):
        # Feed Forward
        dx_1 = self.layernormalization_2.backward(dy)
        dx_2 = self.dropout_2.backward(dx_1)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_FEEDFORWARD)
        dx_2 = self.feedforward.backward(dx_2)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
        dx_2 = dx_1 + dx_2
        # Encoder-Decoder Attention
        dx_2 = self.layernormalization_enc.backward(dx_2)
        # dx_3 = self.dropout_enc.backward(dx_2)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_MHA)
        dx_3, dx_4, dx_5 = self.multiheadattention_enc.backward(dx_3)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
        dx_enc = dx_4 + dx_5
        dx_2 = dx_2 + dx_3
        # Self Attention
        dx_2 = self.layernormalization_1.backward(dx_2)
        # dx_3 = self.dropout_1.backward(dx_2)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_MHA)
        dx_3 = self.multiheadattention.backward(dx_3)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
        # if self.need_dx:
        dx_3, dx_4, dx_5 = dx_3
        dx = dx_2 + dx_3 + dx_4 + dx_5
        return dx, dx_enc
        # else:
        #     return dx_enc



"""PyDTNN environment initialization"""

import os
import atexit
import platform
import subprocess
from types import ModuleType
from collections import Counter

type MPI_MODULE = ModuleType
type Cudnn_Handle_Type = int
type Cublas_Handle_Type = int

gpu_errors = []

# OPTIONAL IMPORTS
try:
    from pydtnn.libs.libmpi import MPI
except Exception:
    MPI = None

try:
    import pycuda.driver as drv  # type: ignore
except Exception as e:
    drv = None
    gpu_errors.append(e)
    has_drv = False
else:
    has_drv = True

try:
    import pycuda.gpuarray as gpuarray  # type: ignore
except Exception as e:
    gpuarray = None
    gpu_errors.append(e)

try:
    from pydtnn.backends.gpu.utils import tensor_gpu  # type: ignore
except Exception as e:
    tensor_gpu = None
    gpu_errors.append(e)

try:
    from pydtnn.libs import libnccl as nccl  # type: ignore
except Exception as e:
    nccl = None
    gpu_errors.append(e)

try:
    from pydtnn.libs import libcudnn as cudnn  # type: ignore
except Exception as e:
    cudnn = None
    gpu_errors.append(e)

try:
    from skcuda import cublas  # type: ignore
except Exception as e:
    cublas = None
    gpu_errors.append(e)


# INIT MPI
if MPI is not None:
    rank = MPI.COMM_WORLD.rank
    nprocs = MPI.COMM_WORLD.size
    hostname = platform.node()
    ranks_per_node = dict(Counter(MPI.COMM_WORLD.allgather(hostname)))
else:
    rank = 0
    nprocs = 1
    hostname = "localhost"
    ranks_per_node = {hostname: nprocs}
# ---

# INIT GPU
try:
    num_gpus = subprocess.check_output(["nvidia-smi", "-L"]).count(b'UUID')
except (FileNotFoundError, subprocess.CalledProcessError):
    num_gpus = 0
os.environ["CUDA_VISIBLE_DEVICES"] = str(rank % num_gpus) if num_gpus else ""
supported_gpu = bool(num_gpus)
# ---

# INIT NCCL
if nccl is not None and num_gpus > 0:
    nccl_id = nccl.ncclGetUniqueId()
    if MPI:
        nccl_id = MPI.COMM_WORLD.bcast(nccl_id)
    nccl_comm = nccl.ncclCommInitRank(nprocs, nccl_id, rank)
    atexit.register(lambda: nccl.ncclCommDestroy(nccl_comm))  # type: ignore
else:
    nccl_comm = None  # type: ignore
# ---

# INIT PYCUDA
if drv is not None:  # equivalent: if has_drv:
    drv.init()
    rank = MPI.COMM_WORLD.rank if MPI else 0
    device = drv.Device(rank % drv.Device.count())
    context = device.make_context()
    stream: drv.Stream = drv.Stream()  # type: ignore
    atexit.register(lambda: context.detach())  # type: ignore
else:
    device = None  # type: ignore
    context = None  # type: ignore
    stream = None  # type: ignore
# ---

# INIT CUDNN
if cudnn is not None and has_drv:
    # NOTE: CUDNN initalization must be done after "drv.init()"
    cudnn_handle: Cudnn_Handle_Type = cudnn.cudnnCreate()  # type: ignore
    atexit.register(lambda: cudnn.cudnnDestroy(cudnn_handle))  # type: ignore
else:
    cudnn_handle: Cudnn_Handle_Type = None  # type: ignore
# ---

# INIT CUBLAS
if cublas is not None:
    cublas_handle: Cublas_Handle_Type = cublas.cublasCreate()  # type: ignore
    atexit.register(lambda: cublas.cublasDestroy(cublas_handle))  # type: ignore
else:
    cublas_handle: Cublas_Handle_Type = None  # type: ignore
# ---

# SYNC CUDNN+PYCUDA
if cudnn is not None and stream is not None:
    cudnn.cudnnSetStream(cudnn_handle, stream.handle)
# ---

# SYNC CUBLAS+PYCUDA
if cublas is not None and stream is not None:
    cublas.cublasSetStream(cublas_handle, stream.handle)  # type: ignore
# ---



"""Message Passing Interface"""

import os as _os
import sys as _sys


# Select implementation
if _os.environ.get("PYMPI_PROTO"):
    from pympi import MPI as _module
else:
    from mpi4py import MPI as _module

# Replace module
_sys.modules[__name__] = _module

if hasattr(_module, "__all__"):
    __all__ = _module.__all__  # type: ignore


def __getattr__(key):
    """Proxy all attributes to module"""
    return getattr(_module, key)



import numpy as np
import pycuda.gpuarray as gpuarray
from pycuda.compiler import SourceModule

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.losses.kl_divergence import KLDivergence
from pydtnn.backends.gpu.losses.loss import LossGPU


class KLDivergenceGPU(KLDivergence[TensorGPU], LossGPU):

    def __init_gpu_kernel__(self):
        module = SourceModule("""
        __global__ void kl_divergence(T *y_targ, T *y_pred, T *res,
                                      T *dx, int b, int bs, int n, T eps)
        {
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            if (idx < b) {
                int i = 0;
                double partial = 0;
                double loss = 0;
                for ( i = 0; i < n; i++ ) {
                    partial = logf(fabs(y_targ[idx * n + i] / (y_pred[idx * n + i] + eps)) + 1.0) / bs;
                    loss += fabs(y_targ[idx * n + i] * partial);
                    dx[idx * n + i] = (T) partial;
                }
                res[idx] = (T) loss;
            }
            return;
        }
        """.replace("T", {np.float32: "float", np.float64: "double"}[self.model.dtype]))
        return module.get_function("kl_divergence")

    def compute(self, y_pred, y_targ, batch_size):
        # loss = SUM(|pred * log(|pred / (targ + eps)| + eps) / N
        # dx = log(|pred / targ + eps| + eps) + 1 / N
        threads = min(self.model.batch_size, 1024)
        blocks = max(self.model.batch_size, 1024) // threads + 1
        self.kernel(y_targ.ary, y_pred.ary, self.loss, self.dx.ary,
                    np.int32(self.model.batch_size), np.int32(batch_size),
                    np.int32(np.prod(self.shape[1:])), np.float32(self.eps),
                    grid=(blocks, 1, 1), block=(threads, 1, 1),
                    stream=self.model.stream)
        # loss = gpuarray.sum(self.loss).get()
        loss = gpuarray.sum(self.dx.ary).get()
        return loss, self.dx



import numpy as np

from pydtnn.layers.layer import Layer

from typing import Any, Callable

from pydtnn.utils.initializers import zeros, ones

from pydtnn.utils.constants import Array, ArrayShape, Parameters

class BatchNormalization[T: Array](Layer[T]):

    def __init__(self, beta=0.0, gamma=1.0, momentum=0.9, epsilon=1e-5,
                 moving_mean_initializer: Callable = zeros,
                 moving_variance_initializer: Callable = ones,
                 sync_stats=False):
        super().__init__()
        self.gamma_init_val = gamma
        self.beta_init_val = beta
        self.momentum = momentum
        self.epsilon = epsilon
        self.moving_mean_initializer: Callable[[ArrayShape, np.dtype], np.ndarray] = moving_mean_initializer
        self.moving_variance_initializer: Callable[[ArrayShape, np.dtype], np.ndarray] = moving_variance_initializer
        self.grad_vars = {Parameters.BETA: Parameters.DBETA, Parameters.GAMMA: Parameters.DGAMMA}
        self.sync_stats = sync_stats
        # The following attributes will be initialized later
        self.co = self.ci = self.hi = self.wi = 0
        self.spatial: bool = None  # type: ignore
        self.gamma: T = None  # type: ignore
        self.beta: T = None  # type: ignore
        self.running_mean: T = None  # type: ignore
        self.running_var: T = None  # type: ignore
        self.std: np.ndarray = None  # type: ignore
        self.xn: np.ndarray = None  # type: ignore
        self.dgamma: T = None  # type: ignore
        self.dbeta: T = None  # type: ignore

    def export(self) -> dict[str, Any]:
        data = super().export()

        data[Parameters.RUNNING_MEAN] = self._export_prop(Parameters.RUNNING_MEAN)
        data[Parameters.RUNNING_VAR] = self._export_prop(Parameters.RUNNING_VAR)

        return data
    # ---

    def import_(self, data: dict[str, Any]) -> None:

        self._import_prop(Parameters.RUNNING_MEAN, data[Parameters.RUNNING_MEAN])
        self._import_prop(Parameters.RUNNING_VAR, data[Parameters.RUNNING_VAR])

        return super().import_(data)

    def initialize(self, prev_shape: ArrayShape, x: T | None):
        super().initialize(prev_shape, x)
        self.shape = prev_shape
        self.spatial = len(self.shape) > 2



"""
Common methods and properties for various unitary tests
"""

import os
import sys
import unittest
import warnings

import numpy as np

from pydtnn.utils import random
from pydtnn.utils.tensor import TensorFormat


# @warning: must be a function, don't use a @property decorator
def verbose_test():
    """Returns True if unittest has been called with -v or --verbose options."""
    return '-v' in sys.argv or '--verbose' in sys.argv


class Params:
    def __init__(self) -> None:
        self.parallel = "sequential"
        self.dtype: np.dtype = np.dtype(np.float32)
        self.tensor_format = TensorFormat.NHWC.upper()
        self.conv_variant = "i2c"

        # model_name = os.environ.get("PYDTNN_TEST_MODEL", "simplecnn")
        # self.model_name = model_name

        self.dataset_name = "synthetic"
        self.synthetic_train_samples = "50000"
        self.synthetic_test_samples = "10000"
        self.synthetic_input_shape = "3,32,32"
        self.synthetic_output_shape = "10"


class TestCase(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        random.seed(0)
        warnings.simplefilter("error")

    def tearDown(self) -> None:
        warnings.resetwarnings()
        super().tearDown()


class D:
    def __init__(self, b=1, c=1, h=128, w=100, kn=1, kh=16, kw=10, vpadding=1, hpadding=1,
                 vstride=1, hstride=1, vdilation=1, hdilation=1):
        self.b = b  # Batch size
        self.c = c  # Channels per layer
        self.h = h  # Layers height
        self.w = w  # Layers width
        self.kn = kn  # Number of filters
        self.kh = kh  # Filters weights height
        self.kw = kw  # Filters weights width
        self.vpadding = vpadding  # Vertical padding
        self.hpadding = hpadding  # Horizontal padding
        self.vstride = vstride  # Vertical stride
        self.hstride = hstride  # Horizontal stride
        self.vdilation = vdilation  # Vertical dilation
        self.hdilation = hdilation  # Horizontal dilation

    @property
    def ho(self):
        return (self.h + 2 * self.vpadding - self.vdilation * (self.kh - 1) - 1) // self.vstride + 1

    @property
    def wo(self):
        return (self.w + 2 * self.hpadding - self.hdilation * (self.kw - 1) - 1) // self.hstride + 1

    @property
    def shape(self):
        return self.b, self.c, self.h, self.w

    def __repr__(self):
        return f"""\
x, weights, and y parameters:
  (b, c, h, w)    = {self.b} {self.c} {self.h} {self.w}
  (kn, c, kh, kw) = {self.kn} {self.c} {self.kh} {self.kw}
  (kn, b, ho, wo) = {self.kn} {self.b} {self.ho} {self.wo}
  padding         = {self.vpadding} {self.hpadding}
  stride          = {self.vstride} {self.hstride}
  dilation        = {self.vdilation} {self.hdilation}
"""


alexnet_layers = [
    # AlexNet Cifar
    D(64, 3, 32, 32, 64, 3, 3, 1, 1, 2, 2, 1, 1),
    D(64, 64, 8, 8, 192, 3, 3, 1, 1, 1, 1, 1, 1),
    D(64, 192, 4, 4, 384, 3, 3, 1, 1, 1, 1, 1, 1),
    D(64, 384, 4, 4, 256, 3, 3, 1, 1, 1, 1, 1, 1),
    D(64, 256, 4, 4, 256, 3, 3, 1, 1, 1, 1, 1, 1),
    # AlexNet ImageNet
    D(64, 3, 227, 227, 96, 11, 11, 1, 1, 4, 4, 1, 1),
    D(64, 96, 27, 27, 256, 5, 5, 1, 1, 1, 1, 1, 1),
    D(64, 256, 13, 13, 384, 3, 3, 1, 1, 1, 1, 1, 1),
    D(64, 384, 13, 13, 384, 3, 3, 1, 1, 1, 1, 1, 1),
    D(64, 384, 13, 13, 256, 3, 3, 1, 1, 1, 1, 1, 1),
]

alexnet_backward_layers = []
for layer in alexnet_layers:
    # w <- y (kn * b * ho * wo)
    alexnet_backward_layers.append(D(layer.c, layer.b, layer.h, layer.w, layer.kn, layer.ho, layer.wo,
                                     layer.vpadding, layer.hpadding, layer.vstride, layer.hstride))

alexnet_all_layers = alexnet_layers + alexnet_backward_layers



"""
Python interface to the PMLib library
"""
import ctypes
import ctypes.util
import functools

import numpy as np

from pydtnn.utils import load_library

_SERVER_IP_LEN = 16
_MAX_TIMING = 10
_LINE_SETSIZE = 16
_N_LINE_BITS = 128


class PMLibServer(ctypes.Structure):
    _fields_ = [("server_ip", ctypes.c_char * _SERVER_IP_LEN),
                ("port", ctypes.c_int)]


class PMLibLines(ctypes.Structure):
    _fields_ = [("__bits", ctypes.c_char * _LINE_SETSIZE)]


class PMLibMeasures(ctypes.Structure):
    _fields_ = [("watts_size", ctypes.c_int),
                ("watts_sets_size", ctypes.c_int),
                ("watts_sets", ctypes.POINTER(ctypes.c_int)),
                ("watts", ctypes.POINTER(ctypes.c_double)),
                ("lines_len", ctypes.c_int)]


class PMLibMeasuresWT(ctypes.Structure):
    _fields_ = [("next_timing", ctypes.c_int),
                ("timing", ctypes.POINTER(ctypes.c_double)),
                ("energy", PMLibMeasures)]


class PMLibCounter(ctypes.Structure):
    _fields_ = [("sock", ctypes.c_int),
                ("aggregate", ctypes.c_int),
                ("lines", PMLibLines),
                ("num_lines", ctypes.c_int),
                ("interval", ctypes.c_int),
                ("measures", ctypes.POINTER(PMLibMeasuresWT))]


class PMLibException(Exception):

    def __init__(self, error):
        self.error = error

    def __str__(self):
        return f'{self.error}'


def check_pmlib_returned_status(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        status = func(*args, **kwargs)
        if status != 0:
            raise PMLibException(f"Call to '{func.__name__}' failed!") from None
        return None

    return wrapper


class PMLib:
    _pmlib = None

    def __init__(self, server_ip, port, verbose=False):
        if self._pmlib is None:
            self._pmlib = load_library("pmlib")
        self.verbose = verbose
        # ----------------
        # Helper functions
        # ----------------
        # int pm_set_server( char *ip, int port, server_t *pm_server);
        self._pmlib.pm_set_server.restype = int
        self._pmlib.pm_set_server.argtypes = [ctypes.c_char_p, ctypes.c_int, ctypes.POINTER(PMLibServer)]
        # int pm_set_lines( char *lines_string ,line_t *lines );
        self._pmlib.pm_set_lines.restype = int
        self._pmlib.pm_set_lines.argtypes = [ctypes.c_char_p, ctypes.POINTER(PMLibLines)]
        # int pm_create_counter(char *pm_id, line_t lines, int aggregate, int interval, server_t pm_server,
        #                       counter_t *pm_counter);
        self._pmlib.pm_create_counter.restype = int
        self._pmlib.pm_create_counter.argtypes = [ctypes.c_char_p, PMLibLines, ctypes.c_int, ctypes.c_int,
                                                  PMLibServer, ctypes.POINTER(PMLibCounter)]
        # int pm_start_counter( counter_t *pm_counter );
        self._pmlib.pm_start_counter.restype = int
        self._pmlib.pm_start_counter.argtypes = [ctypes.POINTER(PMLibCounter)]
        # int pm_stop_counter( counter_t *pm_counter );
        self._pmlib.pm_stop_counter.restype = int
        self._pmlib.pm_stop_counter.argtypes = [ctypes.POINTER(PMLibCounter)]
        # int pm_get_counter_data( counter_t *pm_counter );
        self._pmlib.pm_get_counter_data.restype = int
        self._pmlib.pm_get_counter_data.argtypes = [ctypes.POINTER(PMLibCounter)]
        # int pm_print_data_text(char *file_name,  counter_t pm_counter, line_t lines, int set);
        self._pmlib.pm_print_data_text.restype = int
        self._pmlib.pm_print_data_text.argtypes = [ctypes.c_char_p, PMLibCounter, PMLibLines, ctypes.c_int]
        # int pm_finalize_counter( counter_t *pm_counter );
        self._pmlib.pm_finalize_counter.restype = int
        self._pmlib.pm_finalize_counter.argtypes = [ctypes.POINTER(PMLibCounter)]
        # -----------------------
        # Connect with the server
        # -----------------------
        self.server = PMLibServer()
        self.lines = PMLibLines()
        self.counter = PMLibCounter()
        self.set_server(server_ip, port)
        self.create_lines("0-15")
        # -----------------------
        # Class properties that will be initialized later
        # -----------------------
        self.counter_start_time = None
        self.counter_end_time = None
        self.period = None
        self.len_lines = None
        self.len_samples = None
        self.times = None
        self.watts = None

    def info(self, msg):
        if self.verbose is True:
            print("[PMLib]:", msg)

    @check_pmlib_returned_status
    def set_server(self, server_ip, port):
        self.info("Setting server...")
        return self._pmlib.pm_set_server(server_ip.encode('utf-8'), port, ctypes.byref(self.server))

    @check_pmlib_returned_status
    def create_lines(self, lines_string):
        self.info("Setting lines...")
        return self._pmlib.pm_set_lines(lines_string.encode('utf-8'), ctypes.byref(self.lines))

    @check_pmlib_returned_status
    def create_counter(self, counter_string, aggregate=0, interval=0):
        self.info("Creating counter...")
        return self._pmlib.pm_create_counter(counter_string.encode('utf-8'), self.lines, aggregate, interval,
                                             self.server, ctypes.byref(self.counter))

    @check_pmlib_returned_status
    def start_counter(self):
        self.info("Starting counter...")
        return self._pmlib.pm_start_counter(ctypes.byref(self.counter))

    @check_pmlib_returned_status
    def stop_counter(self):
        self.info("Stopping counter...")
        return self._pmlib.pm_stop_counter(ctypes.byref(self.counter))

    @check_pmlib_returned_status
    def _get_counter_data(self):
        self.info("Getting counter data...")
        return self._pmlib.pm_get_counter_data(ctypes.byref(self.counter))

    @check_pmlib_returned_status
    def print_data_text(self, output_string, set_value):
        self.info(f"Writing data to '{output_string}' file...")
        return self._pmlib.pm_print_data_text(output_string.encode('utf-8'), self.counter, self.lines, set_value)

    @check_pmlib_returned_status
    def finalize_counter(self):
        self.info("Finalizing counter...")
        return self._pmlib.pm_finalize_counter(ctypes.byref(self.counter))

    def get_counter_data(self):
        self._get_counter_data()
        self.counter_start_time, self.counter_end_time = np.ctypeslib.as_array(
            (ctypes.c_double * 2).from_address(ctypes.addressof(self.counter.measures.contents.timing.contents)))
        self.len_lines = 1 if self.counter.aggregate == 1 else self.counter.measures.contents.energy.lines_len
        self.len_samples = self.counter.measures.contents.energy.watts_size // self.len_lines
        self.period = (self.counter_end_time - self.counter_start_time) / (self.len_samples - 1)
        self.times = np.array([self.counter_start_time + x * self.period for x in range(self.len_samples)])
        self.watts = np.ctypeslib \
            .as_array((ctypes.c_double * self.len_samples * self.len_lines).from_address(
                ctypes.addressof(self.counter.measures.contents.energy.watts.contents))) \
            .reshape((self.len_lines, self.len_samples))
        if self.counter.aggregate == 0:
            _sum = np.sum(self.watts, axis=0).reshape(1, -1)
            self.watts = np.concatenate((_sum, self.watts))
            self.len_lines += 1

    def _next_sample_from_start(self, start_time):
        return min(self.len_samples - 1, int((start_time - self.times[0]) / self.period) + 1)

    def _previous_sample_from_end(self, end_time):
        return max(0, int(np.ceil((end_time - self.times[0]) / self.period)) - 1)

    def get_number_of_intermediate_samples(self, start_time, end_time):
        # Next and previous samples from start_time and end_time, respectively
        next_sample_from_start = self._next_sample_from_start(start_time)
        previous_sample_from_end = self._previous_sample_from_end(end_time)
        return max(0, previous_sample_from_end + 1 - next_sample_from_start)

    def get_joules(self, start_time, end_time, debug=False):
        # Check boundaries
        if start_time >= end_time:
            raise ValueError("End time must be greater than start time")
        if start_time < self.times[0]:
            raise ValueError("Given start time is lesser than the counter first time")
        if end_time > self.times[-1]:
            raise ValueError("Given end time is greater than the counter last time")
        # Next and previous samples from start_time and end_time, respectively
        next_sample_from_start = self._next_sample_from_start(start_time)
        previous_sample_from_end = self._previous_sample_from_end(end_time)
        if debug:
            print(f">> {next_sample_from_start=}")
            print(f">> {previous_sample_from_end=}")
        # Interpolate watts for start and end time
        watts_on_start_time, watts_on_end_time = [], []
        for watts in self.watts:
            a, b = np.interp([start_time, end_time], self.times, watts)
            watts_on_start_time.append(a)
            watts_on_end_time.append(b)
        # Promote watts_on_start_time and watts_on_end_time to np.arrays
        watts_on_start_time = np.array(watts_on_start_time)
        watts_on_end_time = np.array(watts_on_end_time)
        if debug:
            print(f">> {watts_on_start_time[0]=} ({self.watts[0, next_sample_from_start]=})")
            print(f">> {watts_on_end_time[0]=} ({self.watts[0, previous_sample_from_end]=})")
        # Integrate the energy
        if next_sample_from_start > previous_sample_from_end:
            # Integrate the energy between the two interpolated samples
            joules = ((watts_on_start_time + watts_on_end_time) / 2) * (end_time - start_time)
        else:
            joules = 0
            # Integrate the energy between start_time and times[next_sample_from_start]
            elapsed_time = self.times[next_sample_from_start] - start_time
            if elapsed_time > 0:
                joules += ((watts_on_start_time + self.watts[:, next_sample_from_start]) / 2) * elapsed_time
            # Integrate the energy between times[previous_sample_from_end] and end_time
            elapsed_time = end_time - self.times[previous_sample_from_end]
            if elapsed_time > 0:
                joules += ((self.watts[:, previous_sample_from_end] + watts_on_end_time) / 2) * elapsed_time
            # Integrate the energy between next_sample_from_start and previous_sample_from_end
            elapsed_time = self.times[previous_sample_from_end] - self.times[next_sample_from_start]
            if elapsed_time > 0:
                joules += np.mean(self.watts[:, next_sample_from_start:previous_sample_from_end + 1], axis=1) \
                    * elapsed_time
        return joules



# Typing related
from typing import List, Dict, Tuple
from pydtnn.layer_base import LayerBase
from pydtnn.activations.activation import Activation
import numpy as np

# Operations/transformations related
import torch
from pydtnn.model import Model as PyDTNN_Model
from pydtnn.layers.input import Input
import pydtnn.converters.pytorch2pydtnn.common as cm
import copy


def load_layers(model: PyDTNN_Model, layers: List[LayerBase], activation_layer: Activation) -> None:
    for layer in layers:
        model.add(layer)
    if not isinstance(layers[-1], Activation) and activation_layer is not None:
        model.add(activation_layer)
    model._initialize()


def extract_layers_relations(model: torch.nn.Module) -> Dict[str, Tuple[str | torch.nn.Module, str]]:
    # TODO: Search the way "torch.fx.symbolic_trace" generates ".code" and not extracting the data from a
    # graph: torch.fx.GraphModule
    graph = torch.fx.symbolic_trace(model)

    # {[output's variable name]: Tuple([string with operation name or the layer object], [string with the args])}
    relations_dic = dict()

    # -- CONSTANTS -- #
    BY_LINES = "\n"
    PSEUDO_INDENTATION = " "
    FIRST_LINE = "forward"
    LAST_LINE = "return"

    SEPARATOR_FUNCTION_VALUE = ";"
    SEPARATOR_ASSIGNATION = " = "
    PARAMETERS_BEGINING = "("
    PARAMETER_ENDING = ")"
    LIST_START = '['
    LIST_SEPARATOR = ','
    LIST_END = ']'
    OPERATION_SEPARATOR = " "  # It is expected that the operator is always between spaces (example: "a + b").

    MODEL_LAYER_REQ = "self"
    MODEL_FUNCT_ARG_NAME = "model"  # NOTE: "model" is the name of the function argument. If it change, it is necessary to change it here.

    TORCH_LAYER_REQ = "torch.nn.functional."
    TORCH_FUNC_REQ = "torch."
    PATTERNS = [TORCH_LAYER_REQ, TORCH_FUNC_REQ]  # NOTE: Order *IS* important.
    # -- END CONSTANTS -- #

    for line in filter(lambda x: not (FIRST_LINE in x or LAST_LINE in x),
                       filter(lambda x: len(x) != 0,
                              [elem.lstrip(PSEUDO_INDENTATION) for elem in graph.code.split(BY_LINES)])):

        # NOTE: seems that there are situations that the line does not have the value.
        line = line.split(SEPARATOR_FUNCTION_VALUE)[0]  # [line, debug's input's value]
        operation = line.split(SEPARATOR_ASSIGNATION)  # [output, function+args]
        if len(operation) > 2:
            # Case: When it is a call to a function with a keyword. Example: "cat = torch.concatenate([var], axis = 1)"
            output_var = operation.pop(0)
            operation = "=".join(operation)  # The spaces are removed to make easier a following step.
        else:
            # Normal case. Example: conv1 = self.conv1(x) ==> operation = [conv1, self.conv1(x)]
            output_var, operation = operation
        # Now we have split the _output's variable_ and the operation.
        # We want to separate the arguments from the function in order to get the layer and the relations with the previous layers.
        operation = operation.split(PARAMETERS_BEGINING)  # [function, ...n..., function, args)]

        func = None  # It will be assigned in the following if-else statement
        if len(operation) > 1:
            # Normal case. Examples: 'getattr(self.layer1, "2").bn1(layer1_2_conv1)', 'self.avgpool(features_36)'
            if any(MODEL_LAYER_REQ in part for part in operation):
                # Case: 'getattr(self.layer1, "2").bn1(layer1_2_conv1)'
                args = operation.pop().replace(PARAMETER_ENDING, "")  # [function, ...n..., function], args
                operation = PARAMETERS_BEGINING.join(operation)  # Reasembling the operation without the arguments.
                operation = operation.replace(MODEL_LAYER_REQ, MODEL_FUNCT_ARG_NAME)
                func = eval(operation)  # Getting the layer object.
            else:
                # Cases: function or layer not defined at model's object's constructor
                # TORCH_LAYER_REQ --> Case: layer not defined at model's object's constructor
                # Example: "adaptive_avg_pool2d = torch.nn.functional.adaptive_avg_pool2d(relu, (1, 1))" ==>
                # ==> operation = "torch.nn.functional.adaptive_avg_pool2d", args = "relu, (1, 1)"
                # NOTE: The first argument is always a previous layer
                # --
                # TORCH_FUNC_REQ --> Case: function. Example: torch.cat()
                _operation = operation.pop(0)  # _operation = something like "torch.cat"; operation= [arg1, arg2) arg3 etc.)] [list[str]]
                args = PARAMETERS_BEGINING.join(operation)[:-1]  # _operation = "torch.cat"; operation= arg1 (arg2) arg3 etc. [str] | [:-1] to remove the final ")"
                operation = _operation

                if operation in cm.SPECIAL_CASES:
                    # TODO [possible future FIXME]: See what to do with the special cases.
                    # continue
                    func = cm.CONCAT  # NOTE: this is a cheap fix. TODO: look what to do in this kind of situations.
                    # "torchvision_models_googlenet_GoogLeNetOutputs": The output is a tuple.
                for pattern in PATTERNS:
                    if pattern in operation:
                        func = operation.replace(pattern, "")  # operation = "adaptive_avg_pool2d" | "cat"
                        break
                    # else: Never happens. One (and only one) of the patterns *must* be in operation
        else:
            # Case "operator". Example; 'layer1_2_bn3 + layer1_1_relu_2'
            # NOTE: It will assumed that *ALWAYS* an operation is between spaces (expected: "3 + l"; unexpected: "3+l").
            #   Also it is asumed that there will be only one operator.
            operation = operation[0].split(OPERATION_SEPARATOR)
            op = operation.pop(1)  # '0:layer1_2_bn3, 1:+, 2:layer1_1_relu_2
            args = ''.join([LIST_START, LIST_SEPARATOR.join(operation), LIST_END])  # '[layer1_2_bn3, layer1_1_relu_2]'
            # args now has the same format as other functions.
            func = cm.switch_operation_symbols(op)
        relations_dic[output_var] = (func, args)
    # end "for line"

    return relations_dic


def convert_layers_and_set_weights_and_biases(input_shape: Tuple[int], layers: Dict[str, Tuple[str | torch.nn.Module, str]]) -> List[LayerBase]:

    converted_layers: Dict[str, LayerBase] = dict()

    # Constants
    # - state_dicts keys.
    LAYER_WEIGHTS = "weight"
    LAYER_BIASES = "bias"

    # - initalizers
    PYDTNN_WEIGHTS_INITIALIZER = "weights_initializer"
    PYDTNN_BIASES_INITIALIZER = "biases_initializer"
    # -
    # -----

    # NOTE: There is no way to get the input shape from a PyTorch model due depends of the dataset ==> The input shape will be a parameter set by the user.
    layer_var_names = list(layers.keys())

    fst_layer = layer_var_names[0]
    _input = layers[fst_layer][1]
    converted_layers[_input] = ((Input(input_shape), None))

    dict_equivalent_layer = dict()
    # If there are two layers like the following ones:
    #   "cat_1 = torch.cat([features_pool0, features_denseblock1_denselayer1_conv2], 1)"
    #   "cat_2 = torch.cat([features_pool0, features_denseblock1_denselayer1_conv2, features_denseblock1_denselayer2_conv2], 1)"
    # features_pool0, features_denseblock1_denselayer1_conv2 are actually "cat_1". The previous dictionary is used to make this "equivalence".

    # layer_var_names: {value's variable (str): ([function (str) or layer (nn.Module)], arguments (str))}
    for operation_variable in layer_var_names:
        operation, params = layers[operation_variable]

        if isinstance(operation, torch.nn.Module):
            layer = operation
            layer_var = operation_variable

            name = layer._get_name()

            # From "vars(layer)" it is possible to get all the information necessary from PyTorch in a dictionary.
            args = {cm.ARGUMENTS: vars(layer)}  # NOTE: In this context, params are the input layers.
            converted_layer = cm.switch_pytorch_pydtnn(name)(args)

            # -- Loading the weigths and the biases into the converted layer -- #
            state_dict = layer.state_dict()
            # There are layers without weight nor biases
            if LAYER_WEIGHTS in state_dict:
                # The weights are "torch.Tensor": torch.Tensor.cpu().detach().numpy() ==> weigths as np.array
                weights: np.ndarray = copy.deepcopy(state_dict[LAYER_WEIGHTS].cpu().detach().numpy())
                # NOTE: There are some layers (like the fully connected) where the shape in PyDTNN is the transpose of the PyTorch's one.
                weights = weights.T if name in cm.TRANSPOSE_WEIGHTS_LAYERS else weights

                if hasattr(converted_layer, PYDTNN_WEIGHTS_INITIALIZER):
                    def weights_initializer(shape: tuple, dtype: np.ndarray, pytorch_weights: np.ndarray = weights, **kwargs_to_ignore) -> np.ndarray:
                        # NOTE [IMPORTANT]: Regarding "pytorch_weights = weights".
                        # NOTE > If "weights" are directly set as the returned value (return weights), for some reason the return will be a reference to "weights"
                        # NOTE >> instead of the "weights" value (that is a reference to the layer's PyTorch's weights), so, since this is in a for loop and
                        # NOTE >>> this function (weights_initializer) is called in some step after the loop, every layer will have the last iteration's "weights" values
                        # NOTE >>>> -a reference to the last layer weights- instead of a reference to their respective layer weights.
                        # NOTE >>>>> In this way "pytorch_weights" has the copy of "weights" values (that, as said before, is a reference to the layer's weights) of that iteration.
                        return pytorch_weights.astype(dtype=dtype, copy=False)
                    # - END weights_initializer - #
                    setattr(converted_layer, PYDTNN_WEIGHTS_INITIALIZER, weights_initializer)
                else:
                    # I'm pretty sure this case never happens (anyways, it's better to have it just in case).
                    converted_layer.weights = weights
            # else: Nothing special.

            if LAYER_BIASES in state_dict:
                biases = copy.deepcopy(state_dict[LAYER_BIASES].cpu().detach().numpy())
                biases = biases.T if name in cm.TRANSPOSE_WEIGHTS_LAYERS else biases
                if hasattr(converted_layer, PYDTNN_BIASES_INITIALIZER):
                    def biases_initializer(shape: tuple, dtype: np.ndarray, pytorch_biases: np.ndarray = biases, **kwargs_to_ignore) -> np.ndarray:
                        # NOTE [IMPORTANT]: See "weights_initializer" notes; the case of "pytorch_biases = biases" parameter is the same case as weights_initializer's "pytorch_weights = weights".
                        return pytorch_biases.astype(dtype=dtype, copy=False)
                    # - END weights_initializer - #
                    setattr(converted_layer, PYDTNN_BIASES_INITIALIZER, biases_initializer)
                else:
                    # As said before, I'm pretty sure this case never happens, but anyways, it's better to have it just in case.
                    converted_layer.biases = biases
            # else: Nothing special.
            # -- END Loading the weigths and the biases into the converted layer -- #

            converted_layers[layer_var] = (converted_layer, params)
        else:  # is intance of string (the name of a function or an operation)
            # Here, params are the input layers and other arguments.
            args = {cm.PARAMETERS: params,
                    cm.LAYERS: converted_layers,
                    cm.EQUIVALENT_LAYERS: dict_equivalent_layer,
                    cm.OPERATION_VAR: operation_variable}

            converted_layers[operation_variable] = cm.function_operation_to_pydtnn(operation)(args)
            # NOTE: Remember, originally these were functions, then they does not have weights nor biases.
    # "for operation_variable in layer_var_names" end.

    list_layers = [layer for layer, _input in converted_layers.values()]
    return list_layers


def check_kwargs_and_set_default(kwargs: dict) -> None:

    DICT_KWARGS_DEFAULT_VALUES = {
        "tensor_format": "NCHW",  # NOTE: PyTorch's weight tensors only NCHW format.
        "model_name": None,  # NOTE: If it's not set to "None", it's possible that other neural network is loaded.
        "batch_size": 64,
        # Model object parameters:
        "omm": None,
        "enable_gpu": False,
        "enable_gpudirect": False,
        "non_blocking_mpi": False,
        "enable_nccl": False,
        "dtype": np.float32,
        "tracing": False,
        "tracer_output": "",
    }

    for k in DICT_KWARGS_DEFAULT_VALUES.keys():
        if k not in kwargs:
            kwargs[k] = DICT_KWARGS_DEFAULT_VALUES[k]


def convert_model(model: torch.nn.Module, input_shape: Tuple[int],
                  default_output_activation_layer: Activation | None = None, **kwargs) -> PyDTNN_Model:
    # "default_output_activation_layer" parameter: if there is no activation layer at the end, the one in this parameter is added to the converted model.
    check_kwargs_and_set_default(kwargs)

    # Output model.
    converted_model = PyDTNN_Model(**kwargs)

    # Obtaining the model's layers/operations, activations, etc.; and the relation between them.
    dict_layers = extract_layers_relations(model=model)

    # Obtaining the PyDTNN equivalent layer for every layer and setting the weights and biases (if it's necessary)
    layers = convert_layers_and_set_weights_and_biases(input_shape=input_shape, layers=dict_layers)

    # Assigning the layers/operations to the converted model and the default activation layer if there is none in the new model.
    load_layers(model=converted_model, layers=layers, activation_layer=default_output_activation_layer)

    return converted_model



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch Linear layers to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict, Any

# Functionality imports
from pydtnn.layers.fc import FC
import pydtnn.converters.pytorch2pydtnn.common as cm


# ------------------ #
def Linear(args: Dict[str, Any]) -> FC:
    # https://pytorch.org/docs/stable/generated/torch.nn.Linear.html#torch.nn.Linear

    # PyTorch attributes:
    # Not used: in_features (It's not used due the way the layer's initialization works in PyDTNN)
    PYTORCH_BIAS = "bias"
    PYTORCH_OUT_FEATURES = "out_features"
    torch_dict_keys = [PYTORCH_BIAS, PYTORCH_OUT_FEATURES]
    # ---- #

    # PyDTNN attributes:
    # Not used: activation
    # Used, but in a different place: weights_initializer, biases_initializer
    PYDTNN_BIAS = "use_bias"
    PYDTNN_SHAPE = "shape"
    pydtnn_dict_keys = [PYDTNN_BIAS, PYDTNN_SHAPE]
    # ---- #

    layer_args = cm.prepare_pydtnn_arguments(arguments=args[cm.ARGUMENTS], torch_dict_keys=torch_dict_keys, pydtnn_dict_keys=pydtnn_dict_keys)

    # PyDTNN expects the shape as a tuple instead of an int.
    if PYDTNN_SHAPE in layer_args and isinstance(layer_args[PYDTNN_SHAPE], int):
        layer_args[PYDTNN_SHAPE] = (layer_args[PYDTNN_SHAPE], )

    return FC(**layer_args)



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def argmax_cython[T: _npDT](x: _npDT_2Dims[T],
                            maxv: _npDT_1Dims[T],
                            amax: np.ndarray[tuple[int], _np.int32],
                            rng: _np.ndarray[tuple[int], _np.int32],
                            axis: int = 0) -> tuple[_npDT_1Dims[T:_np.int32], _npDT_1Dims[T:_np.int32]]:
    """
    Args:
        x (npDT_2Dims): A view 2 dimensional inptu's ndarray.
        maxv (npDT_1Dims): A view to a ndarray of one of the npDT's types where the max values' will be stored.
        amax (np.ndarray[tuple[int], np.int32]): view to a ndarray of type np.int32 where the arg max values' will be stored.
        rng (np.ndarray[tuple[int], np.int32]): view to a ndarray of type np.int32 where some outputs will be stored.
        axis (int): The axis where the argmax will be performed. Can be 0 or 1. Default: 0.

    Returns:
        Explicit: tuple[np.ndarray, np.ndarray]: a tuple formed by: [T: npDT](amax, rng) if axis is 0, or [T: npDT](rng, amax) if not.

        Implicit: maxv.
    """



from __future__ import annotations

import numpy as np

from typing import TYPE_CHECKING, Any
if TYPE_CHECKING:
    from pydtnn.model import Model
    from pydtnn.activations.activation import Activation
    from pydtnn.optimizers.optimizer import Optimizer

from pydtnn import utils
from pydtnn.utils.constants import Array, ArrayShape, Parameters
from pydtnn.backends import PromoteToBackend

try:
    from pycuda.driver import Stream  # type: ignore
except Exception:
    pass


class LayerBase[T: Array](PromoteToBackend):
    def __init__(self, shape: ArrayShape = ()) -> None:
        self.nparams: int = 0
        self.shape: ArrayShape = shape
        self.x: T = None  # type: ignore
        self.y: T = None  # type: ignore
        self.weights: T = None  # type: ignore
        self.biases: T = None  # type: ignore
        self.act: type[Activation] | None = None
        self.grad_vars: dict[str, str] = {}
        self.fwd_time: np.ndarray = None  # type: ignore
        self.bwd_time: np.ndarray = None  # type: ignore
        self.paths: list[list[LayerBase[T]]] = []
        self.reqs_allred = {}
        self.parent_layer: LayerBase | None = None

        # The following attributes will be initialized later
        self.id: int = None  # type: ignore
        self.model: Model = None  # type: ignore
        self.prev_shape: ArrayShape = None  # type: ignore
        self.stream_2: Stream = None  # type: ignore
        self.is_block_layer: bool = False

    @property
    def canonical_name(self) -> str:
        frontend = getattr(self, "_frontend", self)
        return type(frontend).__name__

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def name_with_id(self) -> str:
        return f"{self._id_prefix}{self.name}"

    @property
    def _id_prefix(self) -> str:
        prefix = ''
        if self.id is not None and self.model is not None:
            try:
                model__last_layer = self.model.layers[-1]
            except IndexError:
                max_digits = 1
            else:
                model__last_id = model__last_layer.id
                if len(model__last_layer.children):
                    model__last_id = model__last_layer.children[-1].id
                max_digits = len(str(model__last_id))
            prefix = "{:0{width}d}_".format(self.id, width=max_digits)
        return prefix

    def _show_props(self) -> dict:
        props = {}

        props["id"] = self.id

        paths = []
        curr = self
        while curr.parent_layer is not None:
            for i, path in enumerate(curr.parent_layer.paths):
                for layer in path:
                    if layer.id == curr.id:
                        paths.insert(0, i)
            curr = curr.parent_layer
        if paths:
            props["path"] = ",".join(map(str, paths))

        props["name"] = self.name

        if self.nparams > 0:
            props["params"] = self.nparams
            props["memory"] = utils.convert_size_bytes(self.nparams * self.model.dtype.itemsize)

        if self.prev_shape:
            props["input"] = self.prev_shape

        props["output"] = self.shape

        if len(self.paths) > 0:
            props["paths"] = ", ".join(
                f"{path[0].id}-{path[-1].id}" if path else "Empty"
                for path in self.paths
            )

        if self.weights is not None:
            props["weights"] = self.weights.shape

        return props

    def __repr__(self) -> str:
        props = self._show_props()
        name = props.pop("name")

        props = " ".join(
            f"{key}={value!r}"
            for key, value in props.items()
        )

        return f"<{name} {props}>"

    def initialize(self, prev_shape: ArrayShape, x: T | None = None) -> None:
        self.id = next(self.model.layer_id_generator)
        self.prev_shape = prev_shape
        self.x = x  # type:ignore (If it's used, it will be type "T"; if not, it will never be accesed)
        self.fwd_time = np.zeros((4,), dtype=np.float32)
        self.bwd_time = np.zeros((4,), dtype=np.float32)

    def forward(self, x: T) -> T:
        return x

    def backward(self, dy: T) -> T:
        return dy

    def reduce_weights_async(self, gradient: bool = True):
        pass

    def wait_allreduce_async(self, gradient: bool = True):
        pass

    def reduce_weights_sync(self, gradient: bool = True):
        pass

    def print_in_convdirect_format(self):
        pass

    @property
    def children(self) -> list[LayerBase[T]]:
        children: list[LayerBase[T]] = []
        for path in self.paths:
            children += [layer for layer in path]
        return children

    def update_weights(self, optimizer: Optimizer) -> None:
        optimizer.update(self)

    def _export_prop(self, key: str):
        match key:
            case Parameters.PATHS:
                return [
                    [layer.export() for layer in path]
                    for path in self.paths
                ]

            case _:
                return getattr(self, key, None)

    def _import_prop(self, key: str, value) -> None:
        match key:
            case Parameters.PATHS:
                for layer_path, data_path in zip(self.paths, value):
                    for layer, layer_data in zip(layer_path, data_path):
                        layer.import_(layer_data)

            case _:
                setattr(self, key, value)

    def export(self) -> dict[str, Any]:
        data = {}

        data[Parameters.CANONICAL_NAME] = self._export_prop(Parameters.CANONICAL_NAME)

        for key, value in self.grad_vars.items():
            data[key] = self._export_prop(key)
            data[value] = self._export_prop(value)

        if self.paths:
            data[Parameters.PATHS] = self._export_prop(Parameters.PATHS)

        return data

    def import_(self, data: dict[str, Any]) -> None:
        if data[Parameters.CANONICAL_NAME] != self.canonical_name:
            raise TypeError(f"self type must be the same as the stored data type  (self: {self.canonical_name}, stored: {data[Parameters.CANONICAL_NAME]})")

        for key, value in self.grad_vars.items():
            self._import_prop(key, data[key])
            self._import_prop(value, data[value])

        if Parameters.PATHS in data:
            self._import_prop(Parameters.PATHS, data[Parameters.PATHS])
    # -----


class FusedLayerMixIn[T: Array]():
    def __init__(self, *args, **kwargs):
        from_parent = kwargs.pop("from_parent", None)
        if from_parent is None:
            super().__init__(*args, **kwargs)
        else:
            self.__dict__.update(from_parent)



from typing import Any
import numpy as np
import pycuda.driver as drv  # type: ignore
import pycuda.gpuarray as gpuarray  # type: ignore

from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.model import Model
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.libs import libcudnn as cudnn
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape, Parameters


class BatchNormalizationGPU(BatchNormalization[TensorGPU], LayerGPU):

    @property
    def _ary_prop(self) -> set[str]:
        return {Parameters.RUNNING_MEAN, 
                Parameters.RUNNING_VAR, 
                *super()._ary_prop}
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # NOTE: The next attributes will be initialized later
        self.gamma_beta_mean_var_desc: int | None = None
        self.mode: int = None  # type: ignore
        self.gamma_cpu: np.ndarray = None  # type: ignore
        self.beta_cpu: np.ndarray = None  # type: ignore
        self.dgamma_cpu: np.ndarray = None  # type: ignore
        self.dbeta_cpu: np.ndarray = None  # type: ignore
        self.save_mean: TensorGPU = None  # type: ignore
        self.save_inv_var: TensorGPU = None  # type: ignore
        self.factor: float = None  # type: ignore

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU):
        super().initialize(prev_shape, x)
        self.stream_2 = drv.Stream()

        # Activations y
        y_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.zeros(x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.spatial = len(self.shape) > 2
        self.mode = \
            cudnn.cudnnBatchNormMode['CUDNN_BATCHNORM_SPATIAL' if self.spatial else 'CUDNN_BATCHNORM_PER_ACTIVATION']

        self.gamma_beta_mean_var_desc = cudnn.cudnnCreateTensorDescriptor()
        cudnn.cudnnDeriveBNTensorDescriptor(self.gamma_beta_mean_var_desc,
                                            x.desc, self.mode)
        if self.spatial:
            self.ci, self.hi, self.wi = self.model.decode_shape(prev_shape)
        else:
            self.ci, = prev_shape

        shape_ = (1, self.ci, 1, 1)

        # gamma
        self.gamma_cpu = np.full(shape_, self.gamma_init_val, self.model.dtype)
        gamma_gpu = gpuarray.to_gpu(self.gamma_cpu)
        self.gamma = TensorGPU(gamma_gpu, self.model.tensor_format, self.model.cudnn_dtype)
        # beta
        self.beta_cpu = np.full(shape_, self.beta_init_val, self.model.dtype)
        beta_gpu = gpuarray.to_gpu(self.beta_cpu)
        self.beta = TensorGPU(beta_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.dgamma_cpu, self.dgamma = TensorGPU.initialize(self.gamma.ary.shape, self.model.dtype,
                                                            tensor_format=self.model.tensor_format,
                                                            cudnn_dtype=self.model.cudnn_dtype,
                                                            gpudirect=self.model.gpudirect, 
                                                            drv=(drv if self.model.gpudirect else None))
        self.dbeta_cpu, self.dbeta = TensorGPU.initialize(self.beta.ary.shape, self.model.dtype,
                                                          tensor_format=self.model.tensor_format,
                                                          cudnn_dtype=self.model.cudnn_dtype,
                                                          gpudirect=self.model.gpudirect, 
                                                          drv=(drv if self.model.gpudirect else None))

        running_mean_gpu = gpuarray.to_gpu(self.moving_mean_initializer(shape_, self.model.dtype))
        self.running_mean = TensorGPU(running_mean_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        running_var_gpu = gpuarray.to_gpu(self.moving_variance_initializer(shape_, self.model.dtype))
        self.running_var = TensorGPU(running_var_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        save_mean_gpu = gpuarray.empty(shape_, self.model.dtype)
        self.save_mean = TensorGPU(save_mean_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        save_inv_var_gpu = gpuarray.empty(shape_, self.model.dtype)
        self.save_inv_var = TensorGPU(save_inv_var_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.factor = 1.0 - self.momentum

        self.nparams = self.gamma.size + self.beta.size + self.running_mean.size + self.running_var.size
    # ---

    def forward(self, x: TensorGPU) -> TensorGPU:
        alpha, beta = 1.0, 0.0
        match self.model.mode:
            case Model.Mode.TRAIN:
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
                cudnn.cudnnBatchNormalizationForwardTraining(self.model.cudnn_handle, self.mode,
                                                             alpha, beta, x.desc, x.ptr,
                                                             self.y.desc, self.y.ptr, self.gamma_beta_mean_var_desc,
                                                             self.gamma.ptr,
                                                             self.beta.ptr, self.factor, self.running_mean.ptr,
                                                             self.running_var.ptr,
                                                             self.epsilon, self.save_mean.ptr, self.save_inv_var.ptr)
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
            case Model.Mode.EVALUATE:
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
                cudnn.cudnnBatchNormalizationForwardInference(self.model.cudnn_handle, self.mode,
                                                              alpha, beta, x.desc, x.ptr,
                                                              self.y.desc, self.y.ptr, self.gamma_beta_mean_var_desc,
                                                              self.gamma.ptr,
                                                              self.beta.ptr, self.running_mean.ptr, self.running_var.ptr,
                                                              self.epsilon)
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
            case _:
                raise RuntimeError(f"Unexpected model mode '{self.model.mode}'.")
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        self.x: TensorGPU
        
        alpha_dx, beta_dx, alpha_dgb, beta_dgb = 1.0, 0.0, 1.0, 0.0
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        # Compute dx, dgamma, dbeta
        cudnn.cudnnBatchNormalizationBackward(self.model.cudnn_handle, self.mode,
                                              alpha_dx, beta_dx, alpha_dgb, beta_dgb,
                                              self.x.desc, self.x.ptr, dy.desc, dy.ptr,
                                              self.dx.desc, self.dx.ptr, self.gamma_beta_mean_var_desc,
                                              self.gamma.ptr, self.dgamma.ptr, self.dbeta.ptr, self.epsilon,
                                              self.save_mean.ptr, self.save_inv_var.ptr)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # DtoH dw when data parallelism and no GPU direct/NCCL is used
        if self.model.comm and not self.model.gpudirect and not self.model.enable_nccl:
            self.model.stream.synchronize()
            self.dgamma.ary.get_async(self.stream_2, self.dgamma_cpu)
            self.dbeta.ary.get_async(self.stream_2, self.dbeta_cpu)
        return self.dx
    # -----

    def _export_gamma_beta(self, key: str) -> Any:
        value = getattr(self, key)
        gpu_ary = value.ary
        cpu_ary = gpu_ary.get()
        return np.asarray(np.squeeze(cpu_ary, axis=(0, 2, 3)), dtype=np.float64, order="C", copy=True)
    # ---

    def _export_prop(self, key: str) -> Any:
        match key:
            case Parameters.GAMMA | Parameters.DGAMMA | Parameters.BETA | Parameters.DBETA :
                return self._export_gamma_beta(key)
            case _:
                return super()._export_prop(key)
    # ----

    def _import_gamma_beta(self, key: str, value: Any) -> None:
        attribute = getattr(self, key)
        cpu_ary = np.asarray(np.expand_dims(value, axis=(0, 2, 3)), dtype=self.model.dtype, order="C", copy=None)
        attribute.ary.set(cpu_ary)
        return
    # ---

    def _import_prop(self, key: str, value) -> None:
        match key:
            case Parameters.GAMMA | Parameters.DGAMMA | Parameters.BETA | Parameters.DBETA :
                return self._import_gamma_beta(key, value)
            case _:
                return super()._import_prop(key, value)
    # -----



import numpy as np
import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.compiler import SourceModule  #type: ignore
from pycuda.elementwise import ElementwiseKernel  #type: ignore

from pydtnn.backends.gpu.optimizers.optimizer import OptimizerGPU
from pydtnn.optimizers.nadam import Nadam

from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import DTYPE2CTYPE


class NadamGPU(Nadam[TensorGPU], OptimizerGPU):
    """
    NadamGPU optimizer
    """

    def __init__(self, learning_rate=1e-2, beta1=0.99, beta2=0.999, epsilon=1e-7, decay=0.0, dtype: np.dtype = np.dtype(np.float32)):
        super().__init__(learning_rate, beta1, beta2, epsilon, decay, dtype)

        func_pow = {np.dtype(np.float32): "powf", np.dtype(np.float64): "pow"}
        
        # --- GPU ---
        parameters_gpu = "{T} *w, {T} *dw, {T} *m, {T} *v, float it, " \
                         "float lr, float decay, float beta1, float beta2, float epsilon".format(T=DTYPE2CTYPE[dtype])
        operations_gpu = """
            m[i] = beta1 * m[i] + (1 - beta1) * dw[i];
            v[i] = beta2 * v[i] + (1 - beta2) * {func}(dw[i], 2);
            w[i] -= lr * (decay * w[i] + (((m[i] + (1 - beta1) * dw[i]) / (1 - {func}(beta1, it))) / sqrtf((v[i] / (1 - {func}(beta2, it))) + epsilon)))
        """.format(func=func_pow[dtype])

        self.update_gpu = ElementwiseKernel(parameters_gpu, operations_gpu, "Nadam_kernel")
        # -----------
        
        # GPU DIRECT-
        _name = "Nadam_kernel_gpudirect"
        code = """
            __global__ void {name}({T} *w, {T} *dw, {T} *m, {T} *v,
                                   float it, float lr, float decay,
                                   float beta1, float beta2, float epsilon, int N) 
            {{
                int i = blockIdx.x * blockDim.x + threadIdx.x;
                if (i < N) 
                {{
                    m[i] = beta1 * m[i] + (1 - beta1) * dw[i];
                    v[i] = beta2 * v[i] + (1 - beta2) * {func}(dw[i], 2);
                    w[i] -= lr * (decay * w[i] + (((m[i] + (1 - beta1) * dw[i]) / (1 - {func}(beta1, it))) /
                            sqrt(v[i] / (1 - {func}(beta2, it)) + epsilon)));
                }}
            }}""".format(
                T = DTYPE2CTYPE[dtype],
                func=func_pow[dtype],
                name=_name
            )

        self.update_gpudirect = SourceModule(code).get_function(_name)
        # -----------

    def initialize(self, list_layers: list[LayerGPU]) -> None:
        for layer in list_layers:
            self.context[layer.id] = dict[str, int | gpuarray.GPUArray]()
            self.context[layer.id]["it"] = 0

            for w_ in layer.grad_vars.keys():
                w = getattr(layer, w_)
                self.context[layer.id]["m_%s" % w_] = gpuarray.zeros_like(w.ary, dtype=layer.model.dtype)
                self.context[layer.id]["v_%s" % w_] = gpuarray.zeros_like(w.ary, dtype=layer.model.dtype)

    def update(self, layer: LayerGPU) -> None:
        self.context[layer]["it"] += 1  # type: ignore (self.context[layer]["it"] is always an integer)
        it:int = self.context[layer]["it"]  # type: ignore (self.context[layer]["it"] is always an integer)

        for w_, dw_ in layer.grad_vars.items():
            w, dw = getattr(layer, w_), getattr(layer, dw_)
            m = self.context[layer.id]["m_%s" % w_]
            v = self.context[layer.id]["v_%s" % w_]
            w: TensorGPU
            dw: TensorGPU
            m: gpuarray.GPUArray
            v: gpuarray.GPUArray

            if self.gpudirect:
                n = self.get_batch_size(w)
                threads, blocks = self.get_threads_and_blocks()

                self.update_gpudirect(w.ary.gpudata, dw.ptr_intp, m.gpudata, v.gpudata,
                                      np.float32(it), np.float32(self.learning_rate),
                                      np.float32(self.decay), np.float32(self.beta1),
                                      np.float32(self.beta2), np.float32(self.epsilon),
                                      np.int32(n),
                                      grid=(int(blocks), 1, 1), block=(int(threads), 1, 1),
                                      stream=layer.stream_2)
            else:
                self.update_gpu(w.ary, dw.ary, m, v,
                                np.float32(it), np.float32(self.learning_rate),
                                np.float32(self.decay), np.float32(self.beta1),
                                np.float32(self.beta2), np.float32(self.epsilon),
                                stream=layer.stream_2)



import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.kl_divergence_metric import KLDivergenceMetric


class KLDivergenceMetricCPU(KLDivergenceMetric[np.ndarray], MetricCPU):

    def compute(self, y_pred, y_targ):
        loss = np.abs(y_pred * np.log(np.abs(y_pred / (y_targ + self.eps) + self.eps)))
        loss = np.sum(loss) / y_pred.shape[0]
        return loss



from typing import TYPE_CHECKING
from collections import abc
from contextlib import ExitStack, contextmanager
import io
import copy
import tarfile
from pathlib import Path, PurePath
import typing

import numpy as np
from scipy.io import loadmat

from pydtnn.datasets.dataset import Dataset
from pydtnn.utils import random
from pydtnn.utils.tensor import TensorFormat


if TYPE_CHECKING:
    from pydtnn.model import Model


TRAIN_NSAMPLES = 1281167
TEST_NSAMPLES = 50000
INPUT_SHAPE = (3, 227, 227)
OUTPUT_SHAPE = (1000,)


def list_archive(root_path: Path) -> typing.Iterator[tuple[str, ...]]:
    """Recursive TAR walk"""
    path = str(root_path)

    def is_tar(path: PurePath) -> bool:
        """Does the path look like a TAR"""
        return path.suffix == ".tar"

    if not is_tar(root_path):
        yield (path,)
        return

    with ExitStack() as fp_stack:
        stack: list[tuple[tarfile.TarFile, tuple[str, ...]]] = []

        tar = fp_stack.enter_context(tarfile.open(path, "r:"))
        stack.append((tar, (path,)))

        while stack:
            tar, base_path = stack.pop()

            for member in tar.getmembers():
                if not member.isfile():
                    continue

                path = Path(member.name)
                full_path = (*base_path, str(path))

                if is_tar(path):
                    sub_file = fp_stack.enter_context(tar.extractfile(member))
                    sub_tar = fp_stack.enter_context(tarfile.open(fileobj=sub_file, mode="r:"))
                    stack.append((sub_tar, full_path))
                else:
                    yield full_path


@contextmanager
def load_archive(*paths: str) -> abc.Generator[typing.IO[bytes]]:
    """Recursive TAR loader"""
    with ExitStack() as stack:
        # First: on disk
        file = stack.enter_context(open(paths[0], "rb"))

        if len(paths) <= 1:
            yield file
            return

        tar = stack.enter_context(tarfile.open(fileobj=file, mode="r:"))

        # Intermediate: nested tars
        for fp in paths[1:-1]:
            file = stack.enter_context(tar.extractfile(fp))
            tar = stack.enter_context(tarfile.open(fileobj=file, mode="r:"))

        # Last: return
        file = stack.enter_context(tar.extractfile(paths[-1]))
        yield file


class ImageNet(Dataset):
    """
    ImageNet Dataset

    The most highly-used subset of ImageNet is the ImageNet Large
    Scale Visual Recognition Challenge (ILSVRC) 2012-2017 image
    classification and localization dataset. This dataset spans
    1000 object classes and contains 1,281, 167 training images,
    50,000 validation images and 100,000 test images.

    Source (SHA1): https://image-net.org/challenges/LSVRC/2012/2012-downloads.php
    43eda4fe35c1705d6606a6a7a633bc965d194284 ILSVRC2012_img_train.tar
    5f3f73da3395154b60528b2b2a2caf2374f5f178 ILSVRC2012_img_val.tar
    092a94ed6a05454b8b72d1c4ecf336fa48d37fda ILSVRC2012_devkit_t12.tar.gz

    Normalize (z-score):
    offset: -0.451
    scale:  +3.471
    """

    def __init__(self, model: "Model", force_test_as_validation=False, debug=False):
        super().__init__(model, TRAIN_NSAMPLES, TEST_NSAMPLES, INPUT_SHAPE, OUTPUT_SHAPE, force_test_as_validation=force_test_as_validation, debug=debug)

    def _get_label(self, code: int, labels: dict[int, int]) -> np.ndarray:
        """Transform a code (int) into a label (ndarray 1D uint8)"""
        label = labels[code] - 1
        mask = np.zeros(len(labels), dtype=np.uint8)
        mask[label] = 1
        return mask

    def _get_train_code(self, name: str) -> int:
        """Transform a file-name (str) to code (int)"""
        label = name
        label = label.replace("_", ".")
        label = label.split(".")[0]
        label = label.lstrip("n")
        label = int(label)
        return label

    def _get_val_code(self, name: str) -> int:
        """Transform a file-name (str) to code (int)"""
        label = name
        label = label.replace("_", ".")
        label = label.split(".")[2]
        label = int(label)
        return label

    def _get_train_labels(self, path: Path) -> dict[int, int]:
        """Get label mappings from archive"""
        member = "ILSVRC2012_devkit_t12/data/meta.mat"
        with tarfile.open(path) as tar, tar.extractfile(member) as fp:
            meta = loadmat(file_name=fp, squeeze_me=True)["synsets"]
        nums_children = list(zip(*meta))[4]
        meta = [meta[idx] for idx, num_children in enumerate(nums_children) if num_children == 0]
        labels, codes, class_name = list(zip(*meta))[:3]
        return {
            int(code.lstrip("n")): int(label)
            for code, label in zip(codes, labels)
        }

    def _get_val_labels(self, path: Path) -> dict[int, int]:
        """Get label mappings from archive"""
        member = "ILSVRC2012_devkit_t12/data/ILSVRC2012_validation_ground_truth.txt"
        with tarfile.open(path) as tar, tar.extractfile(member) as fp, io.TextIOWrapper(buffer=fp) as lines:
            return {
                i: int(line)
                for i, line in enumerate(lines, 1)
            }

    def _init_actual_data(self):
        if not self.model.transform_resize:
            raise ValueError("Model transform_resize must be enabled for dataset!")

        meta = Path(self.model.dataset_path) / "ILSVRC2012_devkit_t12.tar.gz"
        train = Path(self.model.dataset_path) / "ILSVRC2012_img_train.tar"
        test = Path(self.model.dataset_path) / "ILSVRC2012_img_val.tar"

        train_lables = self._get_train_labels(meta)

        if len(train_lables) != OUTPUT_SHAPE[0]:
            raise ValueError(f"Mismatch class shape (got: {len(train_lables)}, expect: {OUTPUT_SHAPE[0]})")

        train_xy = [
            (path, self._get_label(self._get_train_code(path[-1]), train_lables))
            for path in list_archive(train)
        ]

        if len(train_xy) != TRAIN_NSAMPLES:
            raise ValueError(f"Mismatch train samples (got: {len(train_xy)}, expect: {TRAIN_NSAMPLES})")

        val_lables = self._get_val_labels(meta)
        val_xy = [
            (path, self._get_label(self._get_val_code(path[-1]), val_lables))
            for path in list_archive(test)
        ]

        if len(val_lables) != TEST_NSAMPLES:
            raise ValueError(f"Mismatch test samples (got: {len(val_lables)}, expect: {TEST_NSAMPLES})")

        self._xy_filenames = [
            train_xy,
            copy.copy(val_xy if self.test_as_validation else train_xy),
            val_xy
        ]

    def _actual_data_generator(self, part):
        offset = self._local_offset[part]
        nsamples = self._local_nsamples[part]
        xy_filenames = self._xy_filenames[part]

        if part is Dataset.Part.TRAIN and self.model.augment_shuffle:
            random.shuffle(xy_filenames)  # type: ignore (numpy shuffle's typing wasn't well defined.)

        xy_filenames = xy_filenames[offset:offset + nsamples]

        for path, y in xy_filenames:
            with load_archive(*path) as fp:
                x = self._load_rgb_image(fp)

            # Add N dimension
            x = x[None, ...]
            y = y[None, ...]

            # Set tensor format
            x = self.model.encode_tensor(x)

            # Set dtype and order
            x = x.astype(dtype=self.model.dtype, order="C")
            y = y.astype(dtype=self.model.dtype, order="C")

            # Inplace normalization
            x /= 255.0

            yield x, y



from functools import partial
from warnings import warn

from pydtnn.backends.cpu.layers.abstract.conv_2d_standard import Conv2DStandardCPU
from pydtnn.libs.libconvdirect import ConvDirect
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.utils.constants import ArrayShape
from pydtnn.utils.tensor import encode_shape

import numpy as np

from pydtnn.utils.tensor import TensorFormat

class Conv2DDirectCPU(Conv2DStandardCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # convDirect related attributes (will be initialized in initialize())
        self.cd = []
    
    def _add_forward_backward_methods(self):
        """Add the different forward and backward methods to the class"""

        def new(name, func):
            func.__name__ = name
            setattr(self, name, func)

        # ConvDirect parameters
        methods = []
        if self.model.conv_direct_method:
            methods = [self.model.conv_direct_method]
        if self.model.enable_best_of:
            if self.model.conv_direct_methods_for_best_of != "":
                methods = self.model.conv_direct_methods_for_best_of.split(',')
        for n, method in enumerate(methods):
            self.cd.append(ConvDirect(method, dtype=self.model.dtype, tensor_format=self.model.tensor_format,
                                        debug=self.debug, parent_layer=self))
            try:
                getattr(self, f"_forward_cd{n}_nhwc")
            except AttributeError:
                new(f"_forward_cd{n}_nhwc", partial(self._forward_cd, n=n))
                new(f"_forward_cd{n}_nchw", partial(self._forward_cd, n=n))
                new(f"_backward_cd{n}_nhwc", partial(self._backward_cd, n=n))
                new(f"_backward_cd{n}_nchw", partial(self._backward_cd, n=n))
    # ----

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)
        self._add_forward_backward_methods()

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                self.forward = self._forward_cd0_nhwc
                self.backward = self._backward_cd0_nhwc
            case TensorFormat.NCHW:
                self.forward = self._forward_cd0_nchw
                self.backward = self._backward_cd0_nchw
            case _:
                raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
        # --

        out_shape = encode_shape((self.model.batch_size, self.co, self.ho, self.wo))
        self.out = np.zeros(out_shape, self.weights.dtype, order="C")

        self.out = None

        if self.use_bias:
            warn(f"\"{self.__class__.__name__}\" never uses the biases.", RuntimeWarning)
    # -----

    def _forward_cd(self, x: np.ndarray, n=0) -> np.ndarray:
        """Version of the forward function that uses the convDirect library"""

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVDIRECT)
        y = self.cd[n].conv_direct(self.weights, x, self.out,
                                   vpadding=self.vpadding, hpadding=self.hpadding,
                                   vstride=self.vstride, hstride=self.hstride,
                                   vdilation=self.vdilation, hdilation=self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return y

    def _backward_cd(self, y: np.ndarray, n=0) -> np.ndarray:
        raise RuntimeError("Backward not implemented yet!")



import numpy as np

from pydtnn.backends.gpu.libs import libcudnn as cudnn

from pydtnn.backends.gpu.layers.abstract.block_layer import AbstractBlockLayerGPU
from pydtnn.layers.multi_head_attention import MultiHeadAttention
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.feed_forward import FeedForward
from pydtnn.layers.layer_normalization import LayerNormalization
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.layers.decoder import Decoder


class DecoderGPU(AbstractBlockLayerGPU, Decoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.multiheadattention = MultiHeadAttention(embedl=self.embedl, d_k=self.d_k, heads=self.heads, dropout_rate=self.dropout_rate)
        # self.dropout_1 = DropoutGPU(rate=self.dropout_rate)
        self.layernormalization_1 = LayerNormalization(axis=(3,))
        self.multiheadattention_enc = MultiHeadAttention(embedl=self.embedl, d_k=self.d_k, heads=self.heads, dropout_rate=self.dropout_rate)
        # self.dropout_enc = DropoutGPU(rate=self.dropout_rate)
        self.layernormalization_enc = LayerNormalization(axis=(3,))
        self.feedforward = FeedForward(shape=(self.embedl,), d_ff=self.d_ff, dropout_rate=self.dropout_rate)
        self.dropout_2 = Dropout(rate=self.dropout_rate)
        self.layernormalization_2 = LayerNormalization(axis=(3,))
        self.paths = [[self.multiheadattention, self.layernormalization_1, self.multiheadattention_enc, self.layernormalization_enc, self.feedforward, self.dropout_2, self.layernormalization_2]]

        # The next attributes will be initialized later
        self.y = self.dx = None

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        x_dec, x_enc, mask_dec = x
        x_dec_shape, x_enc_shape, mask_dec_shape = prev_shape
        mha_shape = (x_dec_shape, mask_dec_shape)

        self.shape = x_enc_shape
        self.first_dims = x_dec.ary.shape[:-1]

        # Initialize all sublayers
        for layer in self.children:
            layer.init_backend_from_model(self.model)

        self.multiheadattention.initialize(prev_shape=mha_shape, x=(x_dec, x_dec, x_dec, mask_dec))
        self.layernormalization_1.initialize(prev_shape=self.shape, x=self.multiheadattention.y)
        self.multiheadattention_enc.initialize(prev_shape=mha_shape, x=(x_dec, x_enc, x_enc, mask_dec))
        self.layernormalization_enc.initialize(prev_shape=self.shape, x=self.multiheadattention_enc.y)

        x_aux = self.flatten(self.layernormalization_enc.y)
        self.layernormalization_enc_y_flatten = TensorGPU(x_aux.ary, self.model.tensor_fmt, self.model.cudnn_dtype)

        self.feedforward.initialize(prev_shape=self.layernormalization_enc.shape, x=self.layernormalization_enc_y_flatten)

        x_aux = self.unflatten(self.feedforward.y)
        self.feedforward_y_unflatten = TensorGPU(x_aux.ary, self.model.tensor_fmt, self.model.cudnn_dtype)
        x_aux = self.unflatten(self.feedforward.dx)
        self.feedforward_dx_unflatten = TensorGPU(x_aux.ary, self.model.tensor_fmt, self.model.cudnn_dtype)

        self.dropout_2.initialize(prev_shape=self.feedforward.shape, x=self.feedforward_y_unflatten)
        x_aux = self.flatten(self.dropout_2.dx)
        self.dropout_2_dx_flatten = TensorGPU(x_aux.ary, self.model.tensor_fmt, self.model.cudnn_dtype)

        self.layernormalization_2.initialize(prev_shape=self.dropout_2.shape, x=self.dropout_2.y)

        self.y = self.layernormalization_2.y
        x_aux = self.multiheadattention.dquery
        self.dx = TensorGPU(x_aux.ary, self.model.tensor_fmt, self.model.cudnn_dtype)
        x_aux = self.multiheadattention_enc.dkey
        self.dx_enc = TensorGPU(x_aux.ary, self.model.tensor_fmt, self.model.cudnn_dtype)
        x_aux = self.multiheadattention_enc.dquery
        self.dx_enc_dquery = TensorGPU(x_aux.ary, self.model.tensor_fmt, self.model.cudnn_dtype)

        self.flatten(self.feedforward.y)
        self.flatten(self.feedforward.dx)

        for layer in self.children:
            self.fwd_time += layer.fwd_time
            self.bwd_time += layer.bwd_time
            self.nparams += layer.nparams

    def initialize_block_layer(self):
        pass

    def flatten(self, x):
        last_dim = x.ary.shape[-1]
        return x.reshape((int(np.prod(self.first_dims)), last_dim))

    def unflatten(self, x):
        last_dim = x.ary.shape[-1]
        return x.reshape((*self.first_dims, last_dim))

    def forward(self, x, x_enc, mask=None):
        alpha, beta = 1.0, 1.0
        # Self Attention
        self.multiheadattention.forward(x, x, x, mask, x)
        self.layernormalization_1.forward(self.multiheadattention.y)

        # Self Attention Encoder
        self.multiheadattention_enc.forward(self.layernormalization_1.y, x_enc, x_enc, mask, self.layernormalization_1.y)
        self.layernormalization_enc.forward(self.multiheadattention_enc.y)

        # Feed Forward
        self.feedforward.forward(self.layernormalization_enc_y_flatten)
        self.dropout_2.forward(self.feedforward_y_unflatten)
        # y = dropout_2.y + layernormalization_enc.y
        cudnn.cudnnAddTensor(self.model.cudnn_handle,
                             alpha, self.layernormalization_enc.y.desc, self.layernormalization_enc.y.ptr,
                             beta, self.dropout_2.y.desc, self.dropout_2.y.ptr)
        self.layernormalization_2.forward(self.dropout_2.y)
        return self.y

    def backward(self, dy):

        alpha, beta = 1.0, 1.0
        # Feed Forward
        self.layernormalization_2.backward(dy)
        self.dropout_2.backward(self.layernormalization_2.dx)
        self.feedforward.backward(self.dropout_2_dx_flatten)

        # dx = feedforward.dx + layernormalization_2.dx
        cudnn.cudnnAddTensor(self.model.cudnn_handle,
                             alpha, self.layernormalization_2.dx.desc, self.layernormalization_2.dx.ptr,
                             beta, self.feedforward_dx_unflatten.desc, self.feedforward_dx_unflatten.ptr)

        # Self Attention Encoder
        self.layernormalization_enc.backward(self.feedforward_dx_unflatten)
        self.multiheadattention_enc.backward(self.layernormalization_enc.dx)
        # dx_enc = multihead.dkey + multiheadattention.dvalue
        cudnn.cudnnAddTensor(self.model.cudnn_handle,
                             alpha, self.dx_enc.desc, self.multiheadattention_enc.dvalue.ptr,
                             beta, self.dx_enc.desc, self.dx_enc.ptr)
        # dx = layernorm_enc.dx + multihead_enc.dquery
        cudnn.cudnnAddTensor(self.model.cudnn_handle,
                             alpha, self.dx_enc_dquery.desc, self.layernormalization_enc.dx.ptr,
                             beta, self.dx_enc_dquery.desc, self.multiheadattention_enc.dquery.ptr)

        # Self Attention
        self.layernormalization_1.backward(self.dx_enc_dquery)
        self.multiheadattention.backward(self.layernormalization_1.dx)
        # if self.need_dx:
        # dx = layernorm_1.dx + multihead.dquery + multihead.dkey + multihead.dvalue
        cudnn.cudnnAddTensor(self.model.cudnn_handle,
                             alpha, self.dx.desc, self.layernormalization_1.dx.ptr,
                             beta, self.dx.desc, self.dx.ptr)
        cudnn.cudnnAddTensor(self.model.cudnn_handle,
                             alpha, self.dx.desc, self.multiheadattention.dkey.ptr,
                             beta, self.dx.desc, self.dx.ptr)
        cudnn.cudnnAddTensor(self.model.cudnn_handle,
                             alpha, self.dx.desc, self.multiheadattention.dvalue.ptr,
                             beta, self.dx.desc, self.dx.ptr)
        return self.dx, self.dx_enc
        # else:
        #     return self.dx_enc
